/**
 * Copyright (C) 1998-2000 by University of Maryland, College Park, MD 20742, USA
 * All rights reserved.
 */
package edu.umd.cs.jazz.event;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;


import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;

/**
 * <b>ZSelectionMoveHandler</b> is a selection handler for use with
 * <b>ZSelectionManager</b>.  <code>ZSelectionMoveHandler</code> allows the user to
 * translate the current selection in two ways:
 * <ul>
 * <li>Any item in the current selection may be dragged with the mouse.  All selected items will be translated.
 * <li>The arrow keys may be used to apply a small translation (or "nudge") to the selected items.
 * </ul>
 * <P>
 * It should be noted that this event handler may not have the desired effect
 * if the event handler is globally active and internal cameras are used.
 * This cannot be fixed in a general purpose way for two reasons: the 
 * possibility of multiple representations and the possibility of multiple
 * cameras viewing the same nodes.  First, an internal camera is a normal
 * visual component and as such can appear in multiple places in a jazz
 * scenegraph.  Consequently, if items in an internal camera are selected
 * and the internal camera appears in multiple places, moving these items 
 * becomes arbitrary (ie. for which representation do they move 
 * appropriately).  Second, if two distinct cameras are viewing the same 
 * set of selected nodes it is again an arbitrary decision as for which 
 * camera the nodes are moved appropriately.  As a result, in the case when
 * the event handler is globally active and internal or multiple cameras
 * appear, the nodes are moved first according to the most recently 
 * interacted camera or, if none has been interacted, according to the
 * top camera (of the ZCanvas).
 * <P>
 * <b>Warning:</b> Serialized and ZSerialized objects of this class will not be
 * compatible with future Jazz releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running the
 * same version of Jazz. A future release of Jazz will provide support for long
 * term persistence.
 *
 * @see edu.umd.cs.jazz.ZSelectionManager
 * @see ZCompositeSelectionHandler
 * @see ZSelectionModifyHandler
 * @see ZSelectionDeleteHandler
 * @see ZSelectionScaleHandler
 *
 * @author Antony Courtney, Yale University
 * @author Lance Good, University of Maryland
 * @author Benjamin Bederson, University of Maryland
 */
public class ZSelectionMoveHandler implements ZEventHandler, ZMouseListener, ZMouseMotionListener, KeyListener, Serializable {
    /** true when event handler is active */
    private boolean active = false;

    /** true if this event handler is active on all cameras */
    private boolean globallyActive = false;

    /** node this event handler attaches to */
    private ZNode node = null;

    /** non-null only if this event handler is active on a single camera */
    private ZCamera activeCamera = null;

    /** the camera with which the event handler is interacting */
    private ZCamera interactionCamera = null;

    /** canvas this event handler attaches to */
    private ZCanvas canvas = null;

    private transient ZNode selNode = null;

    private transient Point2D     pt1, pt2;            // Utiltity points used temporarily within event handlers
    private transient Point2D     pressPt;             // Event coords of mouse press event (in window coordinates)
    private transient Point2D     dragPt;              // Event coords of mouse drag event (in window coordinates)
    private transient Point2D     prevPt;              // Event coords of previous mouse event (in window coordinates)

    private static final int translateLeftKey  = KeyEvent.VK_LEFT;      // Key that translates selected objects left a bit
    private static final int translateRightKey = KeyEvent.VK_RIGHT;     // Key that translates selected objects right a bit
    private static final int translateUpKey    = KeyEvent.VK_UP;        // Key that translates selected objects up a bit
    private static final int translateDownKey  = KeyEvent.VK_DOWN;      // Key that translates selected objects down a bit

    private int allButton1Mask   = (MouseEvent.BUTTON1_MASK | 
				    MouseEvent.ALT_GRAPH_MASK | 
				    MouseEvent.CTRL_MASK | 
				    MouseEvent.SHIFT_MASK);
    
    private int allButton1ButShiftMask   = (MouseEvent.BUTTON1_MASK | 
					    MouseEvent.ALT_GRAPH_MASK | 
					    MouseEvent.CTRL_MASK);


    /**
     * Construct a new ZSelectionMoveHandler.  This event handler will
     * be active on all cameras.
     * @param node The node to which this event handler attaches
     * @param canvas The canvas to which this event handler attaches
     */
    public ZSelectionMoveHandler(ZNode node, ZCanvas canvas) {
	this.node = node;
	this.canvas = canvas;

	pt1 = new Point2D.Double();
	pt2 = new Point2D.Double();
	pressPt = new Point2D.Double();
	prevPt = new Point2D.Double();
	dragPt = new Point2D.Double();

	globallyActive = true;
    }


    /**
     * Construct a new ZSelectionMoveHandler.  This event handler will
     * only be active on the specified camera.
     * @param node The node to which this event handler attaches
     * @param camera The camera to which this event handler attaches
     * @param canvas The canvas to which this event handler attaches
     */
    public ZSelectionMoveHandler(ZNode node, ZCamera camera, ZCanvas canvas) {
	this(node,canvas);
	this.activeCamera = camera;
	globallyActive = false;
    }

    /**
     * Specifies whether this event handler is active
     * @param active True to make this event handler active
     */
    public void setActive(boolean active) {
	if (this.active && !active) {
	    // turn off this handler
	    node.removeMouseListener(this);
	    node.removeMouseMotionListener(this);
	    canvas.removeKeyListener(this);
	    this.active = false;
	} else if (!this.active && active) {
	    // turn on handler:
	    this.active = true;
	    node.addMouseListener(this);
	    node.addMouseMotionListener(this);
	    canvas.addKeyListener(this);
	}
    }

    /**
     * Determines if this event handler is active.
     * @return True if active
     */
    public boolean isActive() {
	return active;
    }

    /**
     * Key press event handler
     * @param <code>e</code> The event.
     */
    public void keyPressed(KeyEvent e) {
	int     keyCode = e.getKeyCode();
	double  panDelta = 1.0; // translation amount for nudge
	double  panX = 0.0, panY = 0.0;
	boolean pan = false;

	switch (keyCode) {
	case translateLeftKey:
	    pan = true;
	    panX = -1.0 * panDelta;
	    break;
	case translateRightKey:
	    pan = true;
	    panX = panDelta;
	    break;
	case translateUpKey:
	    pan = true;
	    panY = -1.0 * panDelta;
	    break;
	case translateDownKey:
	    pan = true;
	    panY = panDelta;
	    break;
	}

	ArrayList selection;
	ZCamera selectionCamera;

	if (interactionCamera != null) {
	    selection = ZSelectionManager.getSelectedNodes(interactionCamera);
	    selectionCamera = interactionCamera;
	}
	else {
	    if (globallyActive) {
		selection = ZSelectionManager.getSelectedNodes(canvas.getCamera());
		selectionCamera = canvas.getCamera();
	    }
	    else {
		selection = ZSelectionManager.getSelectedNodes(activeCamera);
		selectionCamera = activeCamera;
	    }
	}

	if (pan) {
	    double dx, dy;
	    ZNode node;
	    ZTransformGroup transform;
	    // Pan each selected node
	    for (Iterator i=selection.iterator(); i.hasNext();) {
		node = (ZNode)i.next();
		transform = node.editor().getTransformGroup();
				// First convert vector to node coordinate system

		// The best we can do in going from screen to local
		// is cameraToLocal - since a camera might appear
		// in multiple places at multiple scales - there is
		// no way of knowing the actual screenToLocal without
		// a SceneGraphPath (which we aren't guaranteed to have here)

		pt1.setLocation(0.0, 0.0);		
		selectionCamera.cameraToLocal(pt1, transform);
		pt2.setLocation(panX, panY);
		selectionCamera.cameraToLocal(pt2, transform);
		dx = (pt2.getX() - pt1.getX());
		dy = (pt2.getY() - pt1.getY());
				// Then, translate in the node's local coordinate system
		transform.translate(dx, dy);
	    }
	}
    }

    /**
     * Key release event handler
     * @param <code>e</code> The event.
     */
    public void keyReleased(KeyEvent e) {
    }

    /**
     * Key typed event handler
     * @param <code>e</code> The event.
     */
    public void keyTyped(KeyEvent e) {
    }

    /**
     * Mouse press event handler
     * @param <code>e</code> The event.
     */
    public void mousePressed(ZMouseEvent e) {
	if ((e.getModifiers() & allButton1ButShiftMask) == MouseEvent.BUTTON1_MASK) {   // Left button only
	    ZSceneGraphPath path = e.getPath();

	    if (globallyActive) {
		interactionCamera = path.getCamera();
	    }
	    else {
		interactionCamera = activeCamera;
	    }

	    pressPt.setLocation(e.getX(), e.getY());
	    prevPt.setLocation(pressPt);

	    selNode = path.getNode();

	    if (selNode != null) {
		e.consume();
	    }
	}
    }

    /**
     * Mouse drag event handler
     * @param <code>e</code> The event.
     */
    public void mouseDragged(ZMouseEvent e) {
	if ((e.getModifiers() & allButton1ButShiftMask) == MouseEvent.BUTTON1_MASK) {   // Left button only
	    // left button only:
	    ZSceneGraphPath path = e.getPath();

	    dragPt.setLocation(e.getX(), e.getY());

	    // Drag selected objects
	    if (selNode == null) {
		return;
	    }

	    if (interactionCamera != null) {
		double dx, dy;
		ZNode node;
		ZTransformGroup transform;
		ArrayList selection = ZSelectionManager.getSelectedNodes(interactionCamera);
		for (Iterator i=selection.iterator(); i.hasNext();) {
		    node = (ZNode)i.next();
		    transform = node.editor().getTransformGroup();
		    // First convert current and previous point to 
		    // node coordinate systesm
		    pt1.setLocation(prevPt);
		    path.screenToCamera(pt1,interactionCamera);
		    interactionCamera.cameraToLocal(pt1, transform);
		    pt2.setLocation(dragPt);
		    path.screenToCamera(pt2,interactionCamera);
		    interactionCamera.cameraToLocal(pt2, transform);
		    dx = (pt2.getX() - pt1.getX());
		    dy = (pt2.getY() - pt1.getY());
				// Then, translate node
		    transform.translate(dx, dy);
		}
		prevPt.setLocation(dragPt);
		
		e.consume();
	    }
	}
    }

    /**
     * Mouse release event handler
     * @param <code>e</code> The event.
     */
    public void mouseReleased(ZMouseEvent e) {
	if ((e.getModifiers() & allButton1ButShiftMask) == MouseEvent.BUTTON1_MASK) {   // Left button only
	    if (selNode != null) {
		e.consume();
	    }
	}
    }

    /**
     * Invoked when the mouse enters a component.
     */
    public void mouseEntered(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse exits a component.
     */
    public void mouseExited(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse has been clicked on a component.
     */
    public void mouseClicked(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse button has been moved on a node
     * (with no buttons no down).
     */
    public void mouseMoved(ZMouseEvent e) {
    }
}
