/**
 * Copyright (C) 1998-2000 by University of Maryland, College Park, MD 20742, USA
 * All rights reserved.
 */
package edu.umd.cs.jazz.event;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.io.*;
import javax.swing.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;

/**
 * <b>ZoomToEventhandler</b> provides the ability to zoom the current
 * camera to a target node on the drawing surface.
 *
 * On a double click with the right mouse button this event handler will zoom to
 * the node that picked by the mouse click. If no object was picked then it will
 * zoom to its default zoom to node that is specified when the event handler is created
 * or by setDefaultZoomToNode(). The event handler uses two methods to zoom into a target
 * node. The method to use is set by makeZoomToScale(aScale) and makeZoomToScreen(aSpacing).
 *
 * <P>
 * <b>Warning:</b> Serialized and ZSerialized objects of this class will not be
 * compatible with future Jazz releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running the
 * same version of Jazz. A future release of Jazz will provide support for long
 * term persistence.
 *
 * @author  Jesse Grosjean
 */
public class ZoomToEventHandler implements ZEventHandler, ZMouseListener, Serializable {

    private static final double DEFAULT_ZOOM_SCALE = 1.0;
    private static final double DEFAULT_ZOOM_SPACING = 50;
    private static final int DEFAULT_ZOOM_TIME = 500;
    private static final int ZOOM_TO_SCALE = 0;
    private static final int ZOOM_TO_SCREEN = 1;

    private ZNode theNode = null;
    private boolean isActive = false;
    private ZNode theDefaultZoomToNode = null;

    private int theZoomMode = ZOOM_TO_SCREEN;
    private double theZoomScale = DEFAULT_ZOOM_SCALE;
    private double theZoomSpacing = DEFAULT_ZOOM_SPACING;
    private int theZoomTime = DEFAULT_ZOOM_TIME;

    private int allButton1Mask = (MouseEvent.BUTTON1_MASK |
                                  MouseEvent.ALT_GRAPH_MASK |
                                  MouseEvent.CTRL_MASK |
                                  MouseEvent.SHIFT_MASK);

    /**
     * Constructs a new ZoomToEventHandler.
     * @param aNode The node this event handler attaches to.
     */
    public ZoomToEventHandler(ZNode aNode) {
        this(aNode, null);
    }

    /**
     * Constructs a new ZoomToEventHandler.
     * @param aNode The node this event handler attaches to.
     * @param aDefaultZoomToNode The node that we should zoom to if no other node is picked.
     */
    public ZoomToEventHandler(ZNode aNode, ZNode aDefaultZoomToNode) {
        theNode = aNode;
        setDefaultZoomToNode(aDefaultZoomToNode);
    }

    /**
     * Specifies whether this event handler is active or not.
     * @param active True to make this event handler active
     */
    public void setActive(boolean active) {
        if (isActive && !active) {
                                // Turn off event handlers
            isActive = false;
            theNode.removeMouseListener(this);
        } else if (!isActive && active) {
                                // Turn on event handlers
            isActive = true;
            theNode.addMouseListener(this);
        }
    }

    /**
     * Determines if this event handler is active.
     * @return True if active
     */
    public boolean isActive() {
        return isActive;
    }

    /**
     * If the mouse is double clicked then zoom into the picked node.
     * @param <code>e</code> The event.
     */
    public void mousePressed(ZMouseEvent e) {
        if ((e.getModifiers() & allButton1Mask) == MouseEvent.BUTTON1_MASK) {
            if (e.getClickCount() < 2) {
                return;
            }

            zoomTo(e.getPath());
        }
    }

    /**
     * Mouse release event handler
     * @param <code>e</code> The event.
     */
    public void mouseReleased(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse enters a component.
     */
    public void mouseEntered(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse exits a component.
     */
    public void mouseExited(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse has been clicked on a component.
     */
    public void mouseClicked(ZMouseEvent e) {
    }

    /**
     * Get the camera from the scene graph path that we will use to
     * perform the zoom.
     */
    protected ZCamera getCameraToZoomWith(ZSceneGraphPath aPath) {
        return aPath.getCamera();
    }

    /**
     * Get the node the use for the target of the zoom.
     */
    protected ZNode getNodeToZoomTo(ZSceneGraphPath aPath) {
        ZNode result = theDefaultZoomToNode;

        if (aPath.getObject() != null) {
            result = aPath.getNode();
        }

        return result;
    }

    /**
     * Get the amount of time that the zoom animation will take.
     */
    public int getZoomTime() {
        return theZoomTime;
    }

    /**
     * Set the amount of time that the zoom animation will take.
     */
    public int setZoomTime(int aTime) {
        return theZoomTime = aTime;
    }

    /**
     * Set the default node to zoom to. This node is zoomed to when there is
     * a double click, but no object has been picked. Usually the default zoom to
     * node is set to a layer group.
     */
    public void setDefaultZoomToNode(ZNode aNode) {
        theDefaultZoomToNode = aNode;
    }

    /**
     * This method puts the event handler in zoom to scale mode. This means that
     * when the handler zooms to a picked node it will zoom to so that the
     * picked node appears in the specified scale on the screen.
     */
    public void makeZoomToScale(double aScale) {
        theZoomMode = ZOOM_TO_SCALE;
        theZoomScale = aScale;
    }

    /**
     * This method puts the event handler in zoom to screen mode. This means that
     * when the handler zooms to a picked node it will zoom to so that the
     * picked node fills the screen (camera view bounds), minus the specified
     * spacing.
     */
    public void makeZoomToScreen(double aSpacing) {
        theZoomMode = ZOOM_TO_SCREEN;
        theZoomSpacing = aSpacing;
    }

    private void zoomTo(ZSceneGraphPath aPath) {
        ZCamera aCamera = getCameraToZoomWith(aPath);
        ZNode aNode = getNodeToZoomTo(aPath);

        if (aNode == null) return;

        switch (theZoomMode) {
            case ZOOM_TO_SCALE: {
                zoomToScaleWith(aCamera, aNode);
                break;
            }
            case ZOOM_TO_SCREEN: {
                zoomToScreenWith(aCamera, aNode);
                break;
            }
        }
    }

    /**
     * Zoom the specified camera so that the specified target node appears at a certain
     * scale on the screen. This scale is set by makeZoomToScale(aScale);
     */
    protected void zoomToScaleWith(ZCamera aCamera, ZNode aNode) {
        ZBounds nodeBounds = aNode.getGlobalBounds();
        ZBounds cameraBounds = aCamera.getBoundsReference();
        AffineTransform at = new AffineTransform();
        double scale = ZTransformGroup.computeScale(aNode.getGlobalToLocalTransform()) * (theZoomScale/1);
        double ctrX = (0.5 * cameraBounds.getWidth());
        double ctrY = (0.5 * cameraBounds.getHeight());
        double nodeBoundsX = (nodeBounds.getX() + (0.5 * nodeBounds.getWidth()));
        double nodeBoundsY = (nodeBounds.getY() + (0.5 * nodeBounds.getHeight()));
        at.translate(ctrX + (- nodeBoundsX * scale), ctrY + (- nodeBoundsY * scale));
        at.scale(scale, scale);
        aCamera.animate(at, getZoomTime(), aCamera.getDrawingSurface());
    }

    /**
     * Zoom the specified camera so that the specified target node fills the
     * screen minus the spacing set in makeZoomToScreen(aSpacing);
     */
    protected void zoomToScreenWith(ZCamera aCamera, ZNode aNode) {
        ZBounds cameraBounds = aCamera.getBounds();
        cameraBounds.x = cameraBounds.x + theZoomSpacing;
        cameraBounds.y = cameraBounds.y + theZoomSpacing;
        cameraBounds.width = cameraBounds.width - (theZoomSpacing * 2);
        cameraBounds.height = cameraBounds.height - (theZoomSpacing * 2);
        aCamera.center(aNode.getGlobalBounds(), cameraBounds, getZoomTime(), aCamera.getDrawingSurface());
    }
}