

package venice;


import java.util.LinkedList;
import java.util.ListIterator;


/**
 * Class which creates the visualization tree from the
 * GXL model stored in a <code>ModelStorage</code>.
 *
 * The contents of current <code>Visualization</code>
 * <b>are erased</b>, which may or may not be wanted
 * behaviour.
 * 
 * @author	Petteri Kamppuri
 * @version	1.1
 * 
 */
public class CreateVisualizationTreeCommand extends Command
{
	/**
	 * Creates the visualization tree. First all MElements are
	 * added, after that all relationships are added.
	 */
	public void execute()
	{
		mModel		= getAppModule().getModelStorage();
		mVisualization	= getAppModule().getGui().getVisualization();
		
		// We could clear the visualization first.
		mVisualization.clear();
		
                mVisualization.enterNonInteractiveMode();
		addAllModelElements();
		addAllRelationships();
		
		// After the tree has been constructed, we ask the Visualization
		// to make a default layout.
		mVisualization.makeDefaultLayout();

                mVisualization.leaveNonInteractiveMode();
	}
	
	/**
	 * Adds all MElements from <code>ModelStorage</code> to 
	 * current <code>mVisualization</code> instance.
	 * MElements include Components, Interfaces and Packages.
	 * Adding is done by iterating the list of all model nodes.
	 */
	protected void addAllModelElements()
	{
		LinkedList myList = mModel.getRootNodes();
		ListIterator myIterator = myList.listIterator(0);
		
		VPackage myRootPackage = mVisualization.getRootPackage();
		
		// Recursively iterate through MElements and add them to the 
		// visualization tree
		while(myIterator.hasNext())
		{
			Object myObject = myIterator.next();
			
			addModelObjectToVisualizationTree(myRootPackage, myObject);
		}
	}
	
	
	/**
	 * Adds a <code>ModelStorage</code> object and recursively its children
	 * to the visualization tree. Each type of object is treated differently.
	 * Possible types for inModelObject are <code>MPackage</code>,
	 * <code>MComponent</code> and <code>MInterface</code>.
	 * 
	 * @param inVisualizationPackage to what visualization package inModelObject is added
	 * @param inModelObject which model object to add
	 */
	protected void addModelObjectToVisualizationTree(VPackage inVisualizationPackage,
								Object inModelObject)
	{
		if(inModelObject instanceof MPackage)
		{
			MPackage myTempMPackage = (MPackage) inModelObject;
			VPackage myTempVPackage = mVisualization.newVPackage(myTempMPackage.getId());
			
			myTempVPackage.setName(myTempMPackage.getName());
			myTempVPackage.setType(myTempMPackage.getType());
			
			inVisualizationPackage.addVChild(myTempVPackage);
			
			LinkedList myList = mModel.getChildren(myTempMPackage.getId());
			ListIterator myIterator = myList.listIterator(0);
			
			while(myIterator.hasNext())
			{
				addModelObjectToVisualizationTree(myTempVPackage, myIterator.next());
			}
		}
		else if(inModelObject instanceof MComponent)
		{
			MComponent myTempMComponent = (MComponent) inModelObject;
			VComponent myTempVComponent = mVisualization.newVComponent(myTempMComponent.getId());
			
			myTempVComponent.setName(myTempMComponent.getName());
			myTempVComponent.setType(myTempMComponent.getType());
			
			inVisualizationPackage.addVChild(myTempVComponent);
		}
		else if(inModelObject instanceof MInterface)
		{
			MInterface myTempMInterface = (MInterface) inModelObject;
			VInterface myTempVInterface = mVisualization.newVInterface(myTempMInterface.getId());
			
			// Set parameters: name, type and operations
			myTempVInterface.setName(myTempMInterface.getName());
			myTempVInterface.setType(myTempMInterface.getType());
			myTempVInterface.setOperations(myTempMInterface.getOperations().toArray());
			
			inVisualizationPackage.addVChild(myTempVInterface);
		}
		// Should handle errors (unknown object types) here
	}
	
	
	/**
	 * Adds relationships from <code>mModel</code> to <code>mVisualization</code>.
	 * Relationships are fetched as one big linked list and the list is
	 * iterated. For each Object in the list, the method checks if it is an
	 * <code>instanceof MRealization, MGeneralization or MDependency</code>.
	 * If it is, then a corresponding element is added to the visualization.
	 */
	protected void addAllRelationships()
	{
		LinkedList myList;
		ListIterator myIterator;
		
		myList = mModel.getAllRelationships();
		myIterator = myList.listIterator(0);
		
		// Iterate through the list and handle all objects.
		while(myIterator.hasNext())
		{
			Object myObject = myIterator.next();
			
			if(myObject instanceof MDependency)
			{
				MDependency myMDep = (MDependency) myObject;
				VDependency myVDep = mVisualization.newVDependency(myMDep.getId());
				
				myVDep.setDestination(mVisualization.findElementById(myMDep.getDestination()));
				myVDep.setSource(mVisualization.findElementById(myMDep.getSource()));
				myVDep.setType(myMDep.getType());
				
				mVisualization.addVRelationship(myVDep);
				
				//System.out.println("Adding dependency for " + myMDep.getSource());
			}
			else if(myObject instanceof MRealization)
			{
				MRealization myMRealization = (MRealization) myObject;
				VRealization myVRealization = mVisualization.newVRealization(myMRealization.getId());
				
				myVRealization.setDestination(mVisualization.findElementById(myMRealization.getDestination()));
				myVRealization.setSource(mVisualization.findElementById(myMRealization.getSource()));
				myVRealization.setType(myMRealization.getType());
				
				mVisualization.addVRelationship(myVRealization);
				
				//System.out.println("Adding realizations for " + myMRealization.getSource());
			}
			else if(myObject instanceof MGeneralization)
			{
				MGeneralization myMGen = (MGeneralization) myObject;
				VGeneralization myVGen = mVisualization.newVGeneralization(myMGen.getId());
				
				myVGen.setDestination(mVisualization.findElementById(myMGen.getDestination()));
				myVGen.setSource(mVisualization.findElementById(myMGen.getSource()));
				myVGen.setType(myMGen.getType());
				
				mVisualization.addVRelationship(myVGen);
				
				//System.out.println("Adding generalizations for " + myMGen.getSource());
			}
			// Should have an else branch here reporting errors.
		}
		
	}


	/**
	 * Private variable to hold the ModelStorage instance that is needed
	 * in various places in the recursive action of this object. 
	 */
	private ModelStorage  mModel		= null;
	
	/**
	 * Private variable to hold the Visualization instance that is needed
	 * in various places in the recursive action of this object. 
	 */
	private Visualization mVisualization	= null;

}













