package venice;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;

/**
 * Performs loading a GXL model file from an URL.
 * The user is displayed a dialog in which he can
 * select which URL to load.
 * 
 * If the user commits to this action, stream to URL
 * is opened.
 * 
 * @author Petteri Kamppuri
 * @author Hannu Laurila
 */
public class ImportGXLURLCommand extends ImportGXLCommand
{
    private String mSuppliedURL;

    /**
     * constructs a command which tries loading from given URL.
     */
    public ImportGXLURLCommand(String inURL)
    {
	mSuppliedURL = inURL;
    }

    /**
     * constructs a command which asks user for the URL.
     */
    public ImportGXLURLCommand()
    {
	this(null);
    }

    /**
     * Displays a dialog so the user can choose which
     * URL to load.
     * 
     * If the user wanted to load an URL (he can bail out)
     * a stream to URL is opened and that is returned.
     * If open fails or user cancels, null is returned.
     */    
    public InputStream getInputStream()
    {
	String myURLString;

	if (mSuppliedURL == null) {
	    AskUserURLDialog myDialog = 
		new AskUserURLDialog(getAppModule().getGui(),
				     "Load GXL Model from URL", true);
	    
	    myDialog.show();
	    
	    if(myDialog.userWantedToLoad()==false)
		return null;
	    
	    myURLString = myDialog.getURL();
	} else {
	    myURLString = mSuppliedURL;
	}
	
        try {
            URL myURL = new URL(myURLString);
            InputStream myIs = myURL.openStream();
            return myIs;
        }
        catch (IOException myEx) {
            Util.handleException(myEx);
        }

        return null;
    }
    
    /** 
     * Private class to handle user interaction when the 
     * LoadViaHTTPCommand needs to ask the user which URL to load.
     */
    private class AskUserURLDialog extends JDialog implements ActionListener
    {
        /**
         * The only constructor that should be used to instantiate
         * this class. Calls <code>super(Frame, String, boolean)</code>
         * and <code>initialize</code>.
         */
        public AskUserURLDialog(Frame inFrame, 
                                String inName, 
                                boolean inIsModal)
        {
            super(inFrame, inName, inIsModal);
            initialize();
        }
        
            
        /**
         * Inits the dialog with proper items and layout.
         * 
         * <p>Sets the dimensions of the dialog, creates buttons and other
         * items in it and tells them to notify this dialog in case of
         * events.
         */
        protected void initialize()
        {
            mURL = "";
            mUserDidWantToLoad = false;
                
            setBounds(getAppModule().getGui().getBounds().x + 100, 
                      getAppModule().getGui().getBounds().y + 100, 520, 115);
                
            JPanel myPanel    = new JPanel(new FlowLayout());
            JPanel mySecPanel = new JPanel(new FlowLayout());
            JPanel myTriPanel = new JPanel(new BorderLayout());
            JPanel myFouPanel = new JPanel(new BorderLayout());
                
            JLabel myLabel    = new JLabel("URL:");
                
            //myLabel.setPreferredSize(new Dimension(40, 26));
            //myLabel.setMinimumSize(new Dimension(20, 30));
                
            mURLField = new JTextField();
            mURLField.setText(mDefaultURL);
            mURLField.setEditable(true);
            mURLField.setPreferredSize(new Dimension(450, 26));
            mURLField.setMaximumSize(new Dimension(450, 26));
            //mURLField.setMinimumSize(new Dimension(20, 26));
            
            mLoad = new JButton("Load");
            mLoad.addActionListener(this);
            mLoad.setPreferredSize(new Dimension(100, 26));
            mLoad.setMaximumSize(new Dimension(100, 26));
            
            
            mCancel = new JButton("Cancel");
            mCancel.addActionListener(this);
            mCancel.setPreferredSize(new Dimension(100, 26));
            mCancel.setMaximumSize(new Dimension(100, 26));
                
            myPanel.add(mLoad);
            myPanel.add(mCancel);
                
            mySecPanel.add(myLabel);
            mySecPanel.add(mURLField);
                
            myFouPanel.add(myPanel, BorderLayout.EAST);
                
            myTriPanel.add(mySecPanel, BorderLayout.NORTH);		
            myTriPanel.add(myFouPanel, BorderLayout.SOUTH);
                
            getContentPane().add(myTriPanel);
                
            setResizable(false);
        }
            
        /**
         * Returns whether the user wanted to load a model or not.
         * 
         * @return whether the user wanted to load a model
         */
        public boolean userWantedToLoad()
        {
            return mUserDidWantToLoad;
        }
            
            
        /**
         * Returns the URL from which the user wants to load a.
         * GXL model.
         * 
         * @return the URL the user wants to load
         */
        public String getURL()
        {
            return mURL;
        }
            
            
        /**
         * Handles events from buttons.
         * Only events handled now are Load and Cancel
         * buttons.
         * 
         * @param inEvent event which happened
         */
        public void actionPerformed(ActionEvent inEvent)
        {
            Object myEventSource = inEvent.getSource();
                
            if(myEventSource == mLoad) {
                mUserDidWantToLoad = true;
                mURL = mURLField.getText();
                dispose();
            } else if(myEventSource == mCancel) {
                mUserDidWantToLoad = false;
                dispose();
            }
            
        }
        
        /** holds the URL the user has chosen */
        private String		mURL;
        /** used to save information whether the user committed to laoding */
        private boolean		mUserDidWantToLoad;
        /* Swing components */
        private JTextField	mURLField;
        private JButton		mLoad;
        private JButton		mCancel;
        private String          mDefaultURL = 
            "http://www.cs.helsinki.fi/group/venice/smilonen/example.xml";
        

    }
}
