

package venice;


import java.awt.geom.*;


/**
 * Class containing static methods which can be used to calculate
 * intersection points for lines against lines, rectangles and ellipses.
 * 
 * <p>This is highly unoptimized code. Performance can be improved.
 */
public class Intersections
{
	/**
	 * 
	 * Calculates the intersection between two line segments. The intersection
	 * point is returned. If the segments do not intersect, then null is returned.
	 * 
	 * <p><b>Caution!</b> Current implementation does not return an intersection
	 * for two collinear segments. This may be fixed in a later version.
	 * 
	 * <p>Algorithm from comp.graphics.algorithm
	 * <a href="http://www.landfield.com/faqs/graphics/algorithms-faq/">FAQ</a>
	 * 
	 * @param inLine1 first line
	 * @param inLine2 second line
	 * 
	 * @return the intersection point or null if the lines don't intersect
	 * 
	*/
	public static Point2D lineIntersectLine(Line2D inLine1, Line2D inLine2)
	{
		double r;
		double s;
		
		r = 	(	(inLine1.getY1() - inLine2.getY1()) * (inLine2.getX2() - inLine2.getX1()) - 
				(inLine1.getX1() - inLine2.getX1()) * (inLine2.getY2() - inLine2.getY1()) )
			/
			(	(inLine1.getX2() - inLine1.getX1()) * (inLine2.getY2() - inLine2.getY1()) -
				(inLine1.getY2() - inLine1.getY1()) * (inLine2.getX2() - inLine2.getX1()) );
		
		
		s = 	(	(inLine1.getY1() - inLine2.getY1()) * (inLine1.getX2() - inLine1.getX1()) - 
				(inLine1.getX1() - inLine2.getX1()) * (inLine1.getY2() - inLine1.getY1()) )
			/
			(	(inLine1.getX2() - inLine1.getX1()) * (inLine2.getY2() - inLine2.getY1()) -
				(inLine1.getY2() - inLine1.getY1()) * (inLine2.getX2() - inLine2.getX1()) );
		
		
		if(0 <= r && r <= 1 && 0 <= s && s <= 1)
		{
			//System.out.println("Intersects");
			
			Point2D.Double tempPoint = new Point2D.Double(
				inLine1.getX1() + r * (inLine1.getX2() - inLine1.getX1()),
				inLine1.getY1() + r * (inLine1.getY2() - inLine1.getY1())
				);
			
			return tempPoint;
		}
		
		return null;
	}
	
	/**
	 * Calculates whether the given line and rectangle intersect and returns
	 * the intersection point. If they do not intersect, null is returned.
	 * 
	 * <p><b>Caution!</b> A line segment can have two intersection points
	 * with a rectangle. This method returns at most one intersection point.
	 * This method will return the intersection point, which will cut
	 * the line segment to the shortest possible length.
	 * 
	 * @param inLine line segment which to check
	 * @param inRectangle rectangle against which to check intersection
	 * 
	 * @return intersection point with the given rectangle
	 */
	public static Point2D lineIntersectsRectangle(Line2D inLine, RectangularShape inRectangle)
	{
		Point2D	myPoint = null;
		Point2D	myPoint2 = null;
		Point2D	myTempPoint1 = null;
		Point2D	myTempPoint2 = null;
		Line2D	myLine = null;
		
		// NOTE! Always put a value in myPoint first. This function
		// assumes that variable is always not null, when myPoint2 has
		// a value.
		
		// Check for intersection for the first line
		myLine = new Line2D.Double(
					new Point2D.Double(inRectangle.getMinX(), inRectangle.getMinY()),
					new Point2D.Double(inRectangle.getMaxX(), inRectangle.getMinY())
				);
		
		myPoint = lineIntersectLine(inLine, myLine);
		
		// Check for intersection with the second line
		myLine.setLine(
					new Point2D.Double(inRectangle.getMaxX(), inRectangle.getMinY()),
					new Point2D.Double(inRectangle.getMaxX(), inRectangle.getMaxY())
				);
		
		if(myPoint == null)
			myPoint = lineIntersectLine(inLine, myLine);
		else
			myPoint2 = lineIntersectLine(inLine, myLine);
		
		if(myPoint == null || myPoint2 == null)
		{
			// Check for intersection with the third line
			myLine.setLine(
						new Point2D.Double(inRectangle.getMaxX(), inRectangle.getMaxY()),
						new Point2D.Double(inRectangle.getMinX(), inRectangle.getMaxY())
					);
			
			if(myPoint == null)
				myPoint = lineIntersectLine(inLine, myLine);
			else if(myPoint2 == null)
				myPoint2 = lineIntersectLine(inLine, myLine);
			
			if(myPoint == null || myPoint2 == null)
			{
				// Check fourth line
				myLine.setLine(
							new Point2D.Double(inRectangle.getMinX(), inRectangle.getMaxY()),
							new Point2D.Double(inRectangle.getMinX(), inRectangle.getMinY())
						);
				
				if(myPoint == null)
					myPoint = lineIntersectLine(inLine, myLine);
				else if(myPoint2 == null)
					myPoint2 = lineIntersectLine(inLine, myLine);
			}
		}
		
		if(myPoint == null)
		{
			return null;
		}
		else
		{
			// See if myPoint is the only intersection point
			if(myPoint2 == null)
			{
				return myPoint;
			}
			else
			{
				// If there are two intersection points, select the point
				// which causes the line to be the shortest
				double dist1 = myPoint.distanceSq(inLine.getP2());
				double dist2 = myPoint2.distanceSq(inLine.getP2());
				
				if(dist1 < dist2)
					return myPoint;
				else
					return myPoint2;
				
			}
		}
		
		// No intersections, so return null
		// return null;
	}
	
	
	/**
	 * Checks whether the given line intersects with the given ellipse.
	 * Returns the point of intersection or null, if they do not intersect.
	 * A line segment and an ellipse may have two intersection points, but
	 * this method will return only one. There is no guarantee which one
	 * is returned.
	 * 
	 * <p><b>Caution!</b> Currently this code treats the ellipse as a
	 * circle. It assumes it has the same height as width.
	 * 
	 * @param inLine line to check
	 * @param inEllipse ellipse to check against
	 * 
	 * @return point of intersection or null
	 */
	public static Point2D lineIntersectsEllipse(Line2D inLine, RectangularShape inEllipse)
	{
		Point2D myPoint = lineIntersectsRectangle(inLine, inEllipse);

		if(myPoint != null)
		{
			// This is a really ugly way to calculate the intersection point.
			// First assumes the ellipse is a circle. Then it calculates
			// the intersection point to the bounding rectangle and after that
			// scales the point to the arc of the assumed circle.
			
			// Get the circle center
			Point2D myEllipseCenter = new Point2D.Double(inEllipse.getCenterX(), inEllipse.getCenterY());
			
			// Calc vector from center to the rect intersection point
			Point2D myVector = new Point2D.Double(myPoint.getX() - myEllipseCenter.getX(),
								myPoint.getY() - myEllipseCenter.getY());
			
			// Calc length for the vector
			double myVectorLength = myVector.distance(new Point2D.Double(0, 0));
			
			double myEllipseRadius = inEllipse.getWidth() / 2;
			
			// Scale vector from the rectangle edge to the circle
			// NOTE: by changing myEllipseRadius in the Y calc to inEllipse.getHeight() / 2
			// we'd get a real scaling for ellipses
			myVector.setLocation(myVector.getX() * myEllipseRadius / myVectorLength,
						myVector.getY() * myEllipseRadius / myVectorLength);
			
			// Calc new pos from center to the edge to the scaled vector (point = center + vector)
			myPoint.setLocation(myEllipseCenter.getX() + myVector.getX(),
						myEllipseCenter.getY() + myVector.getY());
			
			return myPoint;
		}
		
		return null;
	}
	
	/**
	 * Public main to test this class and its correctness.
	 */
	public static void main(String[] argv)
	{
		Line2D.Double myLine1 = new Line2D.Double(new Point2D.Double(1, 1), new Point2D.Double(6, 2));
		Line2D.Double myLine2 = new Line2D.Double(new Point2D.Double(5, 2), new Point2D.Double(4, 1));
		
		Point2D myTempPoint = lineIntersectLine(myLine1, myLine2);
		
		System.out.println(myTempPoint);
		
		
		Rectangle2D.Double myRect = new Rectangle2D.Double();
		
		myRect.setFrame(4, 0.5, 6, 6);
		
		myTempPoint = lineIntersectsRectangle(myLine1, myRect);
		
		System.out.println(myTempPoint);
		
		
		
		Ellipse2D.Double myEllipse = new Ellipse2D.Double();
		
		myEllipse.setFrame(4, 0.5, 6, 6);
		
		myTempPoint = lineIntersectsEllipse(myLine1, myEllipse);
		
		System.out.println(myTempPoint);
	}
}










