/**
 * @(#) OpenPackageCommand.java
 */

package venice;

import java.util.*;

/**
 * Opens the selected packages. This means that all their children will 
 * become visible and all composite relationships in this package will
 * be opened and the contents of those composite relationships will be
 * put back in the visualization.
 * 
 * @author	Petteri Kamppuri
 * @version	1.2
 */
public class OpenPackageCommand extends Command
{
    /**
     * This method goes through the currently selected nodes from visualization
     * and does the <code>openPackage</code> operation on all <code>VPackage</code>
     * instances.
     */
    public void execute()
    {
	Visualization myVisualization = getAppModule().getGui().getVisualization();
	
	/* Get current selection from visualisation subsystem*/
	VModelElement[] mySelection = myVisualization.getSelection();
	
	if(mySelection != null) {
	    // Iterate through all selected VPackage nodes
	    for(int i = 0; i < mySelection.length; ++i) {
		if(mySelection[i] instanceof VPackage) {
		    openPackage((VPackage) mySelection[i]);
		}
	    }
	}	
    }
    
    /**
     * This method opens the given <code>VPackage</code>.
     * Opening means showing all children and moving all composite relationships
     * down from the package level.
     * 
     * <p>The operation itself is described below as pseudo code.
     *
     * <pre>
     * openPackage(inPackage)
     *         for each inPackage.child as myChild
     *                 myChild.show
     *         iterate inPackage.mCompositeRelationshipList as myElement
     *                 iterate myElement as myRel
     *                         myRel.show
     *                         inPackage.removeFromCompositeRelationships(myRel)
     *         inPackage.setShowContent(true)
     * </pre>
     *
     * @param inPackage the package to open
     */
    public static void openPackage(VPackage inPackage)
    {
	Visualization myVisualization = getAppModule().getGui().getVisualization();
	VModelElement[] myChildren = inPackage.getVChildren();
	
	// Set all children visible
	/*for(int j = 0; j < myChildren.length; ++j) {
	    myChildren[j].setVisible(true);
	    }*/

	inPackage.setShowContent(true);
	
	LinkedList myList = inPackage.getCompositeRelationships();
	
	if(myList != null) {
	    // Iterate through the list of composite relationships,
	    // add each subrelationship to the visualization and remove
	    // the composite one from visualization and the package's
	    // list of composite relationships
	    
	    ListIterator myIterator = myList.listIterator(0);
	    
	    while(myIterator.hasNext()) {
		VRelationship myRel = (VRelationship) myIterator.next();
		
		if(!myRel.getIsRepresentedBySomeoneElse()) {
		    myIterator.remove(); // Remove relationship from this package
		    
		    // Remove relationship from the _other_ end
		    if(myRel.getSource() == inPackage)
			myRel.getDestination().getCompositeRelationships().remove(myRel);
		    else
			myRel.getSource().getCompositeRelationships().remove(myRel);
		    
		    myVisualization.removeVRelationship(myRel);
		    
		    decomposeVRelationship(myRel);
		}
	    }
	}
	
	inPackage.setState(VPackage.OPEN);
    }
    
    /**
     * Decomposes a composite VRelationship. Add all the subrelationships to
     * Visualization
     * 
     * @param inRel the composite relationship to decompose
     */
    public static void decomposeVRelationship(VRelationship inRel)
    {
	Visualization myVisualization = getAppModule().getGui().getVisualization();
	
	myVisualization.removeVRelationship(inRel);
	ListIterator mySubrelIterator = inRel.getSubrelationships().listIterator(0);
	
	while(mySubrelIterator.hasNext()) {
	    VRelationship mySubrelationship = (VRelationship) mySubrelIterator.next();
	    
	    // FIXME: should check for all parents' state up to root package
	    // MEGA FIXME must check for source and destination, states CLOSED and TRANSPARENT
	    /*System.out.println("dest's parent's state " + mySubrelationship.getDestination().getVParent().getState());
	    if(mySubrelationship.getDestination().getVParent().getState() == VPackage.CLOSED) {
		
		VRelationship myNewCompRel = mySubrelationship.getSource().addToCompositeRelationshipList(mySubrelationship);
		mySubrelationship.getSource().debugDump();
		
		if(myNewCompRel != null) {
		    // Add relationship to the _other_ end
		    // FIXME inRel.getSource() was inPackage, should be something...
		    if(myNewCompRel.getSource() == inRel.getSource()) {
			//myRel.getDestination().addToCompositeRelationshipList(myNewCompRel);
			mySubrelationship.getDestination().getCompositeRelationships().add(myNewCompRel);
			System.out.println("Added relationship to the other ends comp rels " + mySubrelationship.getDestination().getId());
		    } else {
			//myRel.getSource().addToCompositeRelationshipList(myNewCompRel);
			mySubrelationship.getSource().getCompositeRelationships().add(myNewCompRel);
			System.out.println("Added relationship to the other ends comp rels 2 " + mySubrelationship.getSource().getId());
		    }
		    System.out.println("myNewCompRel was not null");
		} else {
		    System.out.println("myNewCompRel was null");
		}
		
		mySubrelationship.setIsRepresentedBySomeoneElse(true);
		myVisualization.removeVRelationship(mySubrelationship);
		} else*/ {
		myVisualization.addVRelationshipWithoutAddingOriginalReference(mySubrelationship);
		
		mySubrelationship.setIsRepresentedBySomeoneElse(false);
	    }
	}
    }
}


