/**
 * @(#) VModelElement.java
 */

package venice;

import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.LinkedList;

/**
 * VModelElement is a common interface for different kinds of visualized
 * model elements of which the visualized model consists. Using methods
 * in this interface, common parameters to all visualization elements can
 * be accessed.
 *
 * Every visual model element has an ID which should be unique. Name and
 * type does not have to be unique.
 *
 * @author Hannu Laurila
 */
public interface VModelElement      
{
    /**
     * returns the id of the element
     */
    public String getId();
    
    /**
     * returns the name of the element
     */
    public String getName();
    
    /**
     * returns the type of the element
     */
    public String getType();
    
    /**
     * returns true if element is shown in the visualization
     */
    public boolean getVisible();

    /**
     * sets the name of the element
     */
    public void setName(String inName);
    
    /**
     * sets the type of the element
     */
    public void setType(String inType);
    
    /**
     * sets the visibility of the element
     */
    public void setVisible(boolean inState);

    /**
     * get the point where relationships should be
     * attached to. Probably element center.
     * This method cannot return null.
     *
     * @return the point for relationships to attach to
     */
    public Point2D getRelationshipConnectPoint();
	
    /**
     * Intersect this visualization element with a line. If these graphic elements
     * don't intersect, null is returned.
     *
     * @param inLine the line to calculate intersection with
     * @return the point of intersection or null
     */
    public Point2D intersectWithLine(Line2D inLine);
	
    /**
     * Returns a linked list of relationships that leave
     * the given VPackage and are 
     * attached to this node and are not represented by any other relationship.
     * If this element is a package, relationships from all children that cross
     * the given border are returned also.
     * <p>Relationships can be
     * original relationships as described in the model
     * or temporary relationships created to reprensent
     * composite relationships that are needed when
     * child nodes move their dependencies to their
     * parents.
     * <p>If the parameter is null, all relationships attached to this node
     * are returned.
     * 
     * @param inLeaveThisPackage the package that all returned relationships
     * must leave
     * @return linked list of relationships for this node
     * and its children that cross the given package
     */
    public LinkedList getRelationships(VPackage inLeaveThisPackage);
    
    /**
     * Add the given <code>VRelationship</code>, where this element
     * is either the source or destination, to this element's list of
     * original relationships. Original relationships are those relations
     * that are defined in the GXL model, not created on the fly by
     * the visualization.
     * 
     * @param inRelationship add given relation as an original
     * relationship for this model element.
     */
    public void addRelationship(VRelationship inRelationship);

    /**
     * get parent package for this node.
     */
    public VPackage getVParent();

    /**
     * Returns the list of all composite relationships this element has.
     * 
     * @return the list of composite relationships for this element
     */
    public LinkedList getCompositeRelationships();
    
    /**
     * Add the given relationship to this element's list of composite
     * relationships.
     * 
     * @param inRel The relationship to add to list of composite
     * relationships.
     * @return the new composite VRelationship in which inRel was added or null
     * if inRel was added in an old composite VRelationship
     */
    public VRelationship addToCompositeRelationshipList(VRelationship inRel);

    /**
     * Dump information for this node to stdout.
     */
    public void debugDump();
}
