/**
 * @(#) VRelationship.java
 */

package venice;

import java.util.LinkedList;

/**
 * VRelationship is an visual model element representing an
 * relationship between other two model elements in the architecture.
 * <p>
 * Note that altought it is syntactically possible, you should NOT
 * use another relationships as source or target of a relationship
 *
 * @author Hannu Laurila
 */
public interface VRelationship extends VModelElement
{
    /**
     * Returns the destination of the relationship
     *
     */
    public VModelElement getDestination();
    
    /**
     * Returns the source of the relationship
     */    
    public VModelElement getSource();
    
    /**
     * Sets the destination end of the relationship
     *
     * Arrow end is targeted towards the destination
     */
    public void setDestination(VModelElement inElement);
    
    /**
     * sets the source end of the relationship
     */
    public void setSource(VModelElement inElement);
    
    /**
     * Tells whether this relationship leaves the given <code>VPackage</code>.
     * If inPackage is null, this function returns true.
     * 
     * @param inPackage the package to check for leaving, can be null
     * 
     * @return whether this relationship leaves given package
     */
    public boolean leavesVPackage(VPackage inPackage);
    
    /**
     * Tells whether this VRelationship is a composite relationship
     * representing one or more other relationships. Composite relationships
     * are used when nodes move their dependencies to their parent packages to
     * minimize screen clutter.
     * 
     * @return whether this is a composite relationship
     */
    public boolean isCompositeRelationship();
    
    /**
     * Returns the list of subrelationships if this VRelationship is a
     * composite relationship. If this relationship is not a composite
     * relationship, null is returned.
     * 
     * @return the list of subrelationships for this VRelationship
     */
    public LinkedList getSubrelationships();
    
    /**
     * Add the given VRelationship as a subrelationship for this VRelationship.
     * This function does not check if the given relationships is already in the
     * list. If it is, then that relationship will be twice in the list which is
     * not wanted behaviour. The given relationships is removed from visualization.
     * 
     * @param inRel the relationship to add as subrelationship
     */
    public void addVRelationshipAsSubrelationship(VRelationship inRel);
    
    /**
     * Sets whether this relationship is represented by some other relationship
     * or not.
     * 
     * @param inValue whether this relationship is represented by some other
     * relationship or not
     */
    public void setIsRepresentedBySomeoneElse(boolean inValue);
    
    /**
     * Tells whether this relationship is represented by some other relationship
     * or not.
     * 
     * @return whether this relationship is represented by some other
     * relationship or not
     */
    public boolean getIsRepresentedBySomeoneElse();
    
    /**
     * Removes the given relationship from this relationship's list of
     * subrelationships. If the relationships does not exist in this relationships
     * subrelationship list, nothing happens.
     * 
     * @param inRelationshipToRemove the relationship to remove from 
     * subrelationships
     */
    public void removeSubrelationship(VRelationship inRelationshipToRemove);
    
    /**
     * Set whether relationships draw their labels or not.
     * Initial mode is to have the label invisible.
     * 
     * @param inLabelVisible whether to draw label or not
     */
    public void setLabelVisible(boolean inLabelVisible);
}








