package venice;

import java.io.*;
import java.util.zip.*;

/**
 * VeniceFileHandler is an class which is used to handle loading
 * and saving of .venice-files which is the native storage format
 * of Venice. .venice-files are Zipped files which include both the
 * model in GXL format (model.gxl) and the visualization in
 * visualization subsystems native format (model.visualization).
 * <p>
 * Do not use the same VeniceFileHandler instance for both loading 
 * and saving!
 * <p>
 * Example of using this class for saving a file in .venice-format:
 *
 * <pre>
 * VeniceFileHandler vfh = new VeniceFileHandler();
 *
 * OutputStream gxlStream = vfh.getGXLOutputStream();
 * OutputStream visStream = vfh.getVisualizationOutputStream();
 *
 *        
 * PrintWriter pw;
 *
 * pw = new PrintWriter(gxlStream);
 *
 * pw.println("This is the GXL FILE!");
 * pw.close();
 *
 * pw = new PrintWriter(visStream);
 * pw.println("This is the Visualization file!");
 * pw.close();
 *
 * OutputStream os = new FileOutputStream("testfile.venice");
 * vfh.writeTo(os);
 * </pre>
 * <p>
 * Example of using this class for loading a file in .venice-format:
 *
 * <pre>
 * InputStream is = new FileInputStream("testfile.venice");
 *        
 * vfh = new VeniceFileHandler();
 * vfh.readFrom(is);
 *
 * BufferedReader rdr;
 *
 * rdr = new BufferedReader (new InputStreamReader(vfh.getGXLInputStream()));
 *
 * System.out.println(rdr.readLine());
 * rdr.close();
 *
 * rdr = new BufferedReader (new InputStreamReader(vfh.getVisualizationInputStream()));
 * System.out.println(rdr.readLine());
 * rdr.close();
 * </pre>
 *
 * @author Hannu Laurila
 */
public class VeniceFileHandler
{
    /** size of buffer which is used when reading in contents of the archive */
    private static final int BUFSIZE = 16384;

    /** full contents of the GXL file as a byte array */
    private byte[] mGXLData;

    /** full contents of the visualization file as a byte array */
    private byte[] mVisualizationData;

    /** output stream which is returned to applications wanting to write */
    private ByteArrayOutputStream mGXLOutputStream;

    /** output stream which is returned to applications wanting to write */
    private ByteArrayOutputStream mVisualizationOutputStream;

    /**
     * returns an OutputStream which can be used to read the GXL file
     * embedded in the venice file.
     */
    public OutputStream getGXLOutputStream()
    {
        mGXLOutputStream = new ByteArrayOutputStream();
        return mGXLOutputStream;
    }

    /**
     * returns an OutputStream which can be used to read the Visualization
     * file embeded in the venice file.
     */
    public OutputStream getVisualizationOutputStream()
    {        
        mVisualizationOutputStream = new ByteArrayOutputStream();
        return mVisualizationOutputStream;
    }

    /** 
     * returns an InputStream which can be used to write the GXL file
     * inside the venice file
     */
    public InputStream getGXLInputStream() throws IOException
    {
        if (mGXLData == null)
            throw new IOException("GXL File not available in archive");
        return new ByteArrayInputStream(mGXLData);
    }

    /**
     * returns an InputStream which can be used to write the Visualization
     * file inside the venice file
     */
    public InputStream getVisualizationInputStream() throws IOException
    {
        if (mVisualizationData == null)
            throw new IOException("Visualization file " +
                                  "not available in archive.");
        return new ByteArrayInputStream(mVisualizationData);
    }

    /**
     * reads the venice file from the InputStream to the handler.
     * after this the 'substreams' can be requested using get*InputStream()
     * methods.
     */
    public void readFrom(InputStream ioStream) throws IOException
    {
        ZipInputStream myZipStream = new ZipInputStream(ioStream);
        
        boolean myGXLRead = false;
        boolean myVisRead = false;

        while (myGXLRead == false || myVisRead == false) {
            ZipEntry myEntry = myZipStream.getNextEntry();

            if (myEntry == null)
                break;

            if (myEntry.getName().equalsIgnoreCase("model.gxl")) {
                // Found model.gxl, read it to byte array
                mGXLData = readEntry(myZipStream);
                myZipStream.closeEntry();
                myGXLRead = true;
            }
            if (myEntry.getName().equalsIgnoreCase("model.visualization")) {
                // found model.visualization, read to byte array
                mVisualizationData = readEntry(myZipStream);
                myZipStream.closeEntry();
                myVisRead = true;
            }
        }
        
        myZipStream.close();

        if (myGXLRead == false)
            System.out.println("Archive did not contain GXL file");

        if (myVisRead == false)
            System.out.println("Archive did not contain visualization file");
    }

    /**
     * helper method to read one file from ZipInputStream to a byte buffer.
     */
    private byte[] readEntry(ZipInputStream inZipStream) throws IOException
    {
        ByteArrayOutputStream myBos =
            new ByteArrayOutputStream();
        
        byte[] myTmpArray = new byte[BUFSIZE];
        
        while (true) {
            int count = inZipStream.read(myTmpArray, 0, BUFSIZE);
            if (count < 0)
                break;
            myBos.write(myTmpArray, 0, count);
        }

        return myBos.toByteArray();
    }

    /**
     * Saves the GXL and visualization files associated with this
     * VeniceFileHandler to a venice file (given as an OutputStream)
     */
    public void writeTo(OutputStream ioStream) throws IOException
    {
        if (mGXLOutputStream != null)
            mGXLData = mGXLOutputStream.toByteArray();

        if (mVisualizationOutputStream != null)
            mVisualizationData = mVisualizationOutputStream.toByteArray();

        ZipOutputStream myZipStream = new ZipOutputStream(ioStream);
        myZipStream.setComment("VENICE UML Visualization tool data file");
       
        ZipEntry myEntry;

        if (mGXLData != null) {
            myEntry = new ZipEntry("model.gxl");
            myZipStream.putNextEntry(myEntry);
            myZipStream.write(mGXLData,
                              0,
                              mGXLData.length);
            myZipStream.closeEntry();
        }
        if (mVisualizationData != null) {
            myEntry = new ZipEntry("model.visualization");
            myZipStream.putNextEntry(myEntry);
            myZipStream.write(mVisualizationData, 
                              0, 
                              mVisualizationData.length);
            myZipStream.closeEntry();
        }

        myZipStream.close();
    }

    /**
     * Module test routines.
     */
    public static void main (String[] args) throws Exception
    {
        VeniceFileHandler vfh = new VeniceFileHandler();

        OutputStream gxlStream = vfh.getGXLOutputStream();
        OutputStream visStream = vfh.getVisualizationOutputStream();

        
        PrintWriter pw;

        pw = new PrintWriter(gxlStream);

        pw.println("This is the GXL FILE!");
        pw.close();

        pw = new PrintWriter(visStream);
        pw.println("This is the Visualization file!");
        pw.close();

        OutputStream os = new FileOutputStream("testfile.venice");
        vfh.writeTo(os);

        InputStream is = new FileInputStream("testfile.venice");
        
        vfh = new VeniceFileHandler();
        vfh.readFrom(is);

        BufferedReader rdr;

        rdr = new BufferedReader (new InputStreamReader(vfh.getGXLInputStream()));

        System.out.println(rdr.readLine());
        rdr.close();

        rdr = new BufferedReader (new InputStreamReader(vfh.getVisualizationInputStream()));
        System.out.println(rdr.readLine());
        rdr.close();

        /*
        InputStream is = (new java.net.URL("http://www.japo.fi/~kube/testfile.venice")).openStream();

        System.out.println("Loading from " + is);

        vfh = new VeniceFileHandler();
        vfh.readFrom(is);

        BufferedReader rdr;

        rdr = new BufferedReader (new InputStreamReader(vfh.getGXLInputStream()));

        System.out.println(rdr.readLine());
        rdr.close();
        
        rdr = new BufferedReader (new InputStreamReader(vfh.getVisualizationInputStream()));
        System.out.println(rdr.readLine());
        rdr.close();
        */

    }
}
