/**
 * ZPackage.java
 */

package venice;

import java.awt.*;
import java.awt.geom.*;
import java.util.LinkedList;
import java.util.ListIterator;
import java.io.Serializable;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.component.*;
import edu.umd.cs.jazz.event.*;

/**
 * ZPackage is a concrete implementation of VPackage using Jazz library.
 * ZPackage is implemented as an extension of Jazz's ZGroup which
 * provides useful grouping mechanisms so we do not have to implement
 * them ourselves.
 *
 * @author Hannu Laurila
 */
public class ZPackage extends ZGroup 
    implements VPackage, ZNodeListener, ListenerStateChangeable, Serializable
{
    /** Used for estimating label height on initialization*/
    private static double stAssumedLabelHeight =
	(new ZLabel("DUMMYPACKAGE")).getBoundsReference().getHeight();
    
    // VModelElement instance variables
    
    /** model element id */
    private String mModelElementId;
    
    /** model element name */
    private String mModelElementName;
    
    /** model element type */
    private String mModelElementType;
    
    /** model element visibility */
    private boolean mModelElementVisible;
    
    /** show content (children) of the package */
    private boolean mShowContent;
    
    /**
     * the group below ZPackage which is the real place for all
     * children of type VModelElement. Reason for using another group
     * below the package to store the children is that now we can easily
     * implement fading and content hiding by putting different edit
     * groups above just one child which is mChildGroup.
     */
    private ZGroup mChildGroup;
    
    /** is this package a root package. */
    private boolean mIsRootPackage;
    
    /** visual component used to draw the package label */
    private ZLabel mPackageLabel;
    
    /** visual component used to draw the rectangle around package label */
    private ZRectangle mLabelRectangle;
    
    /** visual component used to draw the rectangle around package contents */
    private ZRectangle mContentRectangle;
    
    /** This packages list of original relationships. */
    private LinkedList mOriginalRelationships = null;
    
    /** This packages list of composite relationships. */
    private LinkedList mCompositeRelationships = null;
    
    /** If package has hidden offered interfaces their substitute
     *  dependencies are kept in this list.
     */
    private LinkedList mOfferedRealizationSubstitutions = null;	
    /** If package has hidden offered interfaces their realizations are 
     *  kept in this list.	
     */
    private LinkedList mHiddenOfferedRealizations = null;
    /** If package has required interfaces drawn their realizations are
     *  kept in this list.
     */
    private LinkedList mRequiredRealizations = null;
    /** 
     * 
     */
    private LinkedList mRequiredInterfaceSubstitutions = null;
    
    /** Package state, OPEN by default.  */
    private int mState = 0;
    
    /**
     * a not-serialized variable telling that no listeners are enabled.
     * after deserialization, the value of this will always be false.
     */
    private transient boolean mListenersEnabled = false;
    

	public ZPackage(String inId, boolean inInitialListenersEnabled)
	{
		this.mModelElementId = inId;
		this.mModelElementName = "";
		this.mModelElementType = "";
		this.mModelElementVisible = true;

		this.mIsRootPackage = false;
		this.mPackageLabel = new ZLabel();
		this.mLabelRectangle = new ZRectangle();
		this.mContentRectangle = new ZRectangle();
		this.mShowContent = true;
                this.mListenersEnabled = inInitialListenersEnabled;

		this.mOfferedRealizationSubstitutions = new LinkedList();
		this.mHiddenOfferedRealizations = new LinkedList();
		this.mRequiredInterfaceSubstitutions = new LinkedList();
		this.mRequiredRealizations = new LinkedList();
		
		this.mCompositeRelationships = new LinkedList();
		
		// FIXME state OPEN by default, remember to change if default state is changed
		this.mState = OPEN;
		
		/* create the content group and add it as a child
		 * BEWARE: addChild causes computeBounds to be called
		 * so first initialize the visual components and then
		 * add the child group.
		 */
		this.mChildGroup = new ZGroup();
		addChild(mChildGroup);

		//mChildGroup.editor().getTransformGroup()
		//.setTranslation(4, 4+stAssumedLabelHeight+4+4);

		setFadingEnabled(true);

		addChildGroupListener();
		
		// if other transform groups are used, put them above before
		// addNodeListener call so the bounds calculations will work
		// correctly

		setChildrenPickable(true);
		setPickable(true);

		dataChanged();
	}

    
    /**
     * Enables fading of package contents while getting closer.
     * Sets up magnification rates which control when fading happens.
     */
    private void setFadingEnabled(boolean inState)
    {
        if (inState == true) {
            ZFadeGroup myFadeGroup = mChildGroup.editor().getFadeGroup();
            myFadeGroup.setFadeType(ZFadeGroup.COMPOSITE_MAG);
            myFadeGroup.setMinMag(0.5);
        } else {
            mChildGroup.editor().removeFadeGroup();
        }
    }



    /** Add relationship to this list.*/
    public void addToHiddenOfferedRealizations(VRelationship inRea)
    {
	mHiddenOfferedRealizations.add(inRea);
    }
    /** Add relationship to this list.*/
    public void addToOfferedRealizationSubstitutions(VRelationship inRea)
    {
	mOfferedRealizationSubstitutions.add(inRea);
    }
    /** Add relationship to this list.*/
    public void addToRequiredInterfaceSubstitutions(VRelationship inRea)
    {
	mRequiredInterfaceSubstitutions.add(inRea);
    }
    /** Add relationship to this list.*/
    public void addToRequiredRealizations(VRelationship inRea)
    {
	mRequiredRealizations.add(inRea);
    }

    /** Clear this list.*/
    public void clearRequiredRealizations()
    {
	mRequiredRealizations.clear();
    }
    /** Clear this list.*/
    public void clearHiddenOfferedRealizations()
    {
	mHiddenOfferedRealizations.clear();
    }
    /** Clear this list.*/
    public void clearOfferedRealizationSubstitutions()
    {
	mOfferedRealizationSubstitutions.clear();
    }
    /** Clear VRelationship inRea from this list.*/
    public void clearRequiredInterfaceSubstitutions(VRelationship inRea)
    {
	mRequiredInterfaceSubstitutions.remove(inRea);
    }
    /** Get content of this list.*/
    public LinkedList getRequiredRealizations()
    {
	return mRequiredRealizations;
    }
    /** Get content of this list.*/
    public LinkedList getHiddenOfferedRealizations()
    {
	return mHiddenOfferedRealizations;
    }
    /** Get content of this list.*/
    public LinkedList getOfferedRealizationSubstitutions()
    {
	return mOfferedRealizationSubstitutions;
    }
    /** Get content of this list.*/
    public LinkedList getRequiredInterfaceSubstitutions()
    {
	return mRequiredInterfaceSubstitutions;
    }


    private void removeChildGroupListener()
    {
	if (mListenersEnabled)
	    mChildGroup.editor().getTop().removeNodeListener(this);
    }
    
    private void addChildGroupListener()
    {
	if (mListenersEnabled)
	    mChildGroup.editor().getTop().addNodeListener(this);
    }
    
    public void setListenersEnabled(boolean inEnabled, boolean inRecurse)
    {
        if (mListenersEnabled == true && inEnabled == false)
            mChildGroup.editor().getTop().removeNodeListener(this);        
        else if (mListenersEnabled == false && inEnabled == true)
            mChildGroup.editor().getTop().addNodeListener(this);

        mListenersEnabled = inEnabled;
        
        if (inRecurse) {
            VModelElement[] myChilds = getVChildren();
            for (int i=0; i < myChilds.length; i++) {
                VModelElement myElem = myChilds[i];
                if (myElem instanceof ListenerStateChangeable)
                    ((ListenerStateChangeable)myElem).
                        setListenersEnabled(inEnabled, inRecurse);
            }
        }        
        
        dataChanged();
    }

	public void addVChild(VModelElement inElem)
	{
            mChildGroup.addChild((ZNode) inElem);
            dataChanged();
	}

	public boolean getShowContent()
	{
		return this.mShowContent;
	}

	public VModelElement[] getVChildren()
	{
		ZNode[] myChildNodes =
		    mChildGroup.getChildren();
		VModelElement[] myElems =
		    new VModelElement[mChildGroup.getNumChildren()];

		for (int i=0; i < myChildNodes.length; i++) {
			ZNode myNode = myChildNodes[i];
			ZNode myRealNode = myNode.editor().getNode();

                        // FIXME: this check should not be needed.
                        if (myRealNode instanceof VModelElement) {
                            myElems[i] = (VModelElement) myRealNode;
                        } else {
                            System.err.println("ZPackage.getVChildren: myRealNode is " +
                                               "not a VModelElement. Check scenegraph.");
                            System.err.println("\tZPackage: " + this.toString());
                            System.err.println("\tMyNode: " + myNode.toString());
                            System.err.println("\tMyRealNode: " + myRealNode.toString());
                            System.err.println("\tReturned array will include a null pointer!");
                        }                        
		}

		return myElems;
	}

    public void removeVChild(VModelElement inElem)
    {
	if (inElem == null)
	    return;
	
	ZNode[] myChildNodes =
	    mChildGroup.getChildren();
	
	for (int i=0; i < myChildNodes.length; i++) {
	    ZNode myNode = myChildNodes[i];
	    ZNode myRealNode = myNode.editor().getNode();
	    // FIXME: is that check below ok?
            // should we maybe event myRealNode.editor().getNode() ???
	    if(myRealNode instanceof VModelElement) {
		VModelElement myFoundElem = (VModelElement) myRealNode;
		if (myFoundElem == inElem) {
                    // found the element to be removed. remove the node.
		    mChildGroup.removeChild(myNode);
		    break;
		}
            } else {
                System.err.println("ZPackage.removeVChild: myRealNode is " +
                                   "not a VModelElement. Check scenegraph.");
                System.err.println("\tZPackage: " + this.toString());
                System.err.println("\tMyNode: " + myNode.toString());
                System.err.println("\tMyRealNode: " + myRealNode.toString());
            }
	}
    }

    public VPackage getVParent()
    {
	ZNode myPackage = this;

	// go up the tree until we find a ZPackage.

	do {
	    myPackage = myPackage.getParent();
	} while ( (myPackage != null) && !(myPackage instanceof ZPackage) );

	if (myPackage != null && myPackage instanceof ZPackage)
	    return (VPackage) myPackage;
	else
	    return null;
    }
    
	public void setShowContent(boolean inState)
	{
		if (inState == mShowContent)
			// no change!
			return;

		mShowContent = inState;

		removeChildGroupListener();

		if (mShowContent == false)
			mChildGroup.editor().getInvisibleGroup();
		else
			mChildGroup.editor().removeInvisibleGroup();

		addChildGroupListener();
		
		dataChanged();
	}

	/**
	 * creates the default layout for the package.
	 *
	 * Layout making is either a recursive or a non-recursive operation.
	 * If recursion is used, children ZPackages are laid out before
	 * laying out this ZPackage.
	 *
	 * @param inRecurse if true, recurses to children ZPackages.
	 *
	 */
	public void makeDefaultLayout(boolean inRecurse, int level)
	{
		removeChildGroupListener();

		VModelElement[] myChildren = getVChildren();

		// first layout the children packages of this packages - then
		// this package itself

		// recurse to children if requested
		if (inRecurse == true) {
			// process children recursively
			for (int i = 0; i < myChildren.length; i++)
				if (myChildren[i] instanceof ZPackage)
					((ZPackage)myChildren[i]).makeDefaultLayout(true, level+1);
		}

		// layout this package using grid layout.

                // count number of non-VRelationship-children
                // so we can decided row and column count.

                int myLayoutCount = 0;

                for (int i=0; i < myChildren.length; i++)
                    if (! (myChildren[i] instanceof VRelationship))
                        myLayoutCount++;
                
		int myChildsPerRow = (int) Math.sqrt(myLayoutCount);
                
		int myChildsInThisRow = 0;

		// start so that whole package's (topx,topy) will be about at (0,0)
		// FIXME: this calculation might not be right at all
		double myThisRowY = 4 + stAssumedLabelHeight + 4 + 4;
		double myThisRowX = 4 + 4;
		double myThisRowHeight = 0;

		final double FREESPACE = 40.0;

		for (int i = 0; i < myChildren.length; i++) {
			VModelElement myChild =  myChildren[i];

			// System.out.println("MyChild: " + myChild);

			// don't layout relationships, they won't obey anyway.
			if (myChild instanceof ZRelationship)
				continue;

			ZNode myNode = (ZNode) myChild;

			ZBounds myChildBounds = myNode.getBoundsReference();

			myNode.editor().getTransformGroup()
			.setTranslation(myThisRowX, myThisRowY);

			myThisRowX += myChildBounds.getWidth() + FREESPACE;
			if (myThisRowHeight < myChildBounds.getHeight())
				myThisRowHeight = myChildBounds.getHeight();

			myChildsInThisRow++;

			if (myChildsInThisRow == myChildsPerRow) {
				// proceed to next row
				myThisRowY += myThisRowHeight + FREESPACE;
				myThisRowX = 0;
				myChildsInThisRow = 0;
				myThisRowHeight = 0;
			}
		}

		// set scaling according to the level

		mChildGroup.editor().getTransformGroup().scale(1.0/Math.sqrt(level));

		addChildGroupListener();

		this.dataChanged();
	}

	/**
	 * computes the bounds of this ZPackage. It is a union of
	 * content rectangle and the label rectangle. In case of root package,
	 * we use normal ZGroup bounds calculation
	 */
	protected void computeBounds()
	{
		if (mIsRootPackage) {
			super.computeBounds();
		} else {
			bounds.reset();
			bounds.add(mContentRectangle.getBoundsReference());
			bounds.add(mLabelRectangle.getBoundsReference());
		}
	}

	/**
	 * renders this package
	 */
	public void render(ZRenderContext ctx)
	{
		// first render the rectangles, then label and then children

		// if in root package, don't render label and rectangles
		if (mIsRootPackage == false) {
			mLabelRectangle.render(ctx);
			mContentRectangle.render(ctx);
			mPackageLabel.render(ctx);
		}

		super.render(ctx);
	}

	public boolean pick(Rectangle2D rect, ZSceneGraphPath path)
	{
		// System.err.println("Path on enter: " + path);

		// try picking children. if picked, return.
		if (super.pick(rect,path) == true)
			return true;

		// if we are a root package, this ZPackage can't be picked so
		// return

		if (mIsRootPackage)
			return false;

		// try picking label and content rectangles. if they pick, return
		// as if this ZPackage was picked.

		if (mLabelRectangle.pick(rect,path) ||
		        mContentRectangle.pick(rect,path)) {
			path.setObject(this);
			// System.err.println("Path on exit: " + path);
			return true;
		}


		return false;
	}

	/**
	 * modify the package rectangles etc. called when for example
	 * child moves so we need to determine new positions for the
	 * rectangles
	 */
	protected void dataChanged()
	{
		if (mModelElementVisible == false)
			this.editor().getInvisibleGroup();
		else
			this.editor().removeInvisibleGroup();

		// if we are root package, don't bother adjusting rectangles
		// - just reshape

		if (mIsRootPackage) {
			reshape();
			return;
		}

		this.mPackageLabel.setText(mModelElementName);

		ZBounds myChildBounds =
		    mChildGroup.editor().getTop().getBoundsReference();
		ZBounds myLabelBounds =
		    mPackageLabel.getBoundsReference();

		double myContentRectX = myChildBounds.getX() - 4;
		double myContentRectY = myChildBounds.getY() - 4;
		double myContentRectW = 4 + myChildBounds.getWidth() + 4;
		double myContentRectH = 4 + myChildBounds.getHeight() + 4;

		double myLabelX = myChildBounds.getX();
		double myLabelY = myContentRectY - 4 - myLabelBounds.getHeight();

		double myLabelRectX = myContentRectX;
		double myLabelRectY = myLabelY - 4;
		double myLabelRectW = 4 + myLabelBounds.getWidth() + 4;
		double myLabelRectH = 4 + myLabelBounds.getHeight() + 4;

		mPackageLabel.setTranslation(myLabelX, myLabelY);

		mLabelRectangle.setRect(myLabelRectX, myLabelRectY,
		                        myLabelRectW, myLabelRectH);

		// ensure that content rectangle is >=20 units bigger than labelrect

		if (myContentRectW < (myLabelRectW + 20))
			myContentRectW = myLabelRectW + 20;

		// make sure that content rectangle is at least two times labelrect

		if (myContentRectH < (2*myLabelRectH))
			myContentRectH = 2 * myLabelRectH;

		mContentRectangle.setRect(myContentRectX, myContentRectY,
		                          myContentRectW, myContentRectH);

		reshape();
	}

	/**
	 * Notified when childrens bounds change. Causes dataChanged()
	 * to be called.
	 */
	public void boundsChanged(ZNodeEvent e)
	{
		dataChanged();
	}

	public void globalBoundsChanged(ZNodeEvent e)
	{
		// we are not interested in global bounds changes here.
	}

	/**
	 * clears the package. Not part of VPackage interface.
	 *
	 * @param inRecurse if true, clears the children ZPackages too.
	 */
	public void removeAllVChildren(boolean inRecurse)
	{
		VModelElement[] myChildren = getVChildren();

		for (int i = 0; i < myChildren.length; i++) {
			if ((myChildren[i] instanceof ZPackage) &&
			        (inRecurse == true))
				((ZPackage)myChildren[i]).removeAllVChildren(true);
			removeVChild(myChildren[i]);
		}
	}

	/**
	 * set if this package is the root package. For root package
	    * the label and rectangles are now drawn.
	    *
	    * @param inIsRootPackage true, if this is root package
	    */
	public void setRootPackage(boolean inIsRootPackage)
	{
		mIsRootPackage = inIsRootPackage;

		removeChildGroupListener();
		setFadingEnabled(!inIsRootPackage);
		addChildGroupListener();

		dataChanged();
	}

	/**
	 * determine if this package is root package.
	 *
	 * @return true if root package
	 */
	public boolean isRootPackage()
	{
		return mIsRootPackage;
	}

	// VModelElements are duplicated here because we can't have
	// a common superclass with 'leafs'

	// VModelElement operations begin

	public String getId()
	{
		return this.mModelElementId;
	}

	public String getName()
	{
		return this.mModelElementName;
	}

	public String getType()
	{
		return this.mModelElementType;
	}

	public boolean getVisible()
	{
		return this.mModelElementVisible;
	}

	public void setName(String inName)
	{
		if (inName == null)
			inName = "";

		this.mModelElementName = inName;
		this.dataChanged();
	}

	public void setType(String inType)
	{
		if (inType == null)
			inType = "";

		this.mModelElementType = inType;
		this.dataChanged();
	}

	public void setVisible(boolean inState)
	{
		if (inState == this.mModelElementVisible)
			// no change
			return;

		this.mModelElementVisible = inState;
		this.dataChanged();
	}

	public Point2D getRelationshipConnectPoint()
	{
		ZBounds myBounds = getGlobalBounds();

		return myBounds.getCenter2D();
	}

	public Point2D intersectWithLine(Line2D inLine)
	{
		Rectangle2D myContentBounds = mContentRectangle.getBounds();
		Rectangle2D myLabelBounds = mLabelRectangle.getBounds();

		localToGlobal(myContentBounds);
		localToGlobal(myLabelBounds);

		Point2D myContentIntersectionPoint =
		    Intersections.lineIntersectsRectangle(inLine, myContentBounds);
		Point2D myLabelIntersectionPoint =
		    Intersections.lineIntersectsRectangle(inLine, myLabelBounds);

		if(myContentIntersectionPoint == null
		        && myLabelIntersectionPoint == null)
			return null;

		if(myContentIntersectionPoint == null
		        && myLabelIntersectionPoint != null)
			return myLabelIntersectionPoint;

		if(myContentIntersectionPoint != null
		        && myLabelIntersectionPoint == null)
			return myContentIntersectionPoint;

		// Use squared distances for greater performance
		double myContentPointDistance =
		    myContentIntersectionPoint.distanceSq(inLine.getP2());
		double myLabelPointDistance =
		    myLabelIntersectionPoint.distanceSq(inLine.getP2());

		if(myContentPointDistance < myLabelPointDistance)
			return myContentIntersectionPoint;
		else
			return myLabelIntersectionPoint;
	}

	/**
	 * Returns a list of relationships which start from this package
	 * and span across the border of the given package and are not
	 * represented by some other relationship.
	 */
	public LinkedList getRelationships(VPackage inCrossThisPackage)
	{
	        LinkedList	myTempList = new LinkedList();
		ListIterator	myIterator;
		
		if(inCrossThisPackage == this)
		{
			if(mOriginalRelationships != null) {
			    ListIterator myTempIterator = mOriginalRelationships.listIterator(0);
			    
			    while(myTempIterator.hasNext()) {
				VRelationship myTempRelationship = (VRelationship)
				    myTempIterator.next();
				
				if(!myTempRelationship.getIsRepresentedBySomeoneElse())
				    myTempList.add(myTempRelationship);
			    }
			}
			if(mCompositeRelationships != null) {
			    ListIterator myTempIterator = mCompositeRelationships.listIterator(0);
			    
			    while(myTempIterator.hasNext()) {
				VRelationship myTempRelationship = (VRelationship)
				    myTempIterator.next();
				
				if(!myTempRelationship.getIsRepresentedBySomeoneElse()) {
				    System.out.println("Added composite relationship");
				    myTempList.add(myTempRelationship);
				}
			    }					
			}
			if(!mOfferedRealizationSubstitutions.isEmpty()){
			    myIterator = mOfferedRealizationSubstitutions.listIterator(0);
			    
			    while(myIterator.hasNext()){
				VRelationship myRel = (VRelationship) myIterator.next();
				if(myRel.leavesVPackage(inCrossThisPackage) &&
				   !myRel.getIsRepresentedBySomeoneElse())
				    {
					myTempList.add(myRel);
				    }
			    }
			}
			if(!mRequiredInterfaceSubstitutions.isEmpty()){
			    myIterator = mRequiredInterfaceSubstitutions.listIterator(0);
			    
			    while(myIterator.hasNext())
				{
				    VRelationship myRel = (VRelationship) myIterator.next();
				    if(myRel.leavesVPackage(inCrossThisPackage) &&
				       !myRel.getIsRepresentedBySomeoneElse())
					{
					    myTempList.add(myRel);
					}
				    }
			}
			
		}
		else
		{
			if(mOriginalRelationships != null)
			    {
				myIterator = mOriginalRelationships.listIterator(0);
				
				while(myIterator.hasNext())
				    {
					VRelationship myRel = (VRelationship) myIterator.next();
					
					if(myRel.leavesVPackage(inCrossThisPackage) &&
					   !myRel.getIsRepresentedBySomeoneElse())
					    {
						myTempList.add(myRel);
					    }
				    }
			    }
			if(mCompositeRelationships != null)
			    {
				myIterator = mCompositeRelationships.listIterator(0);
				
				while(myIterator.hasNext())
				    {
					VRelationship myRel = (VRelationship) myIterator.next();
					
					if(myRel.leavesVPackage(inCrossThisPackage) &&
					   !myRel.getIsRepresentedBySomeoneElse())
					    {
						myTempList.add(myRel);
					    }
				    }
			    }
			
                        if(!mOfferedRealizationSubstitutions.isEmpty())
			    {
				myIterator = mOfferedRealizationSubstitutions.listIterator(0);
				
				while(myIterator.hasNext())
				    {
                                        VRelationship myRel = (VRelationship) myIterator.next();
					if(myRel.leavesVPackage(inCrossThisPackage) &&
                                           !myRel.getIsRepresentedBySomeoneElse())
                                            {
						myTempList.add(myRel);
                                            }
				    }
			    }
			if(!mRequiredInterfaceSubstitutions.isEmpty())
                            {
				myIterator = mRequiredInterfaceSubstitutions.listIterator(0);

				while(myIterator.hasNext())
				    {
					VRelationship myRel = (VRelationship) myIterator.next();
					if(myRel.leavesVPackage(inCrossThisPackage) &&
                                           !myRel.getIsRepresentedBySomeoneElse())
					    {
						myTempList.add(myRel);
					    }
                                    }
			    }
		}
		
		// Iterate children
		VModelElement[] myChildList = getVChildren();
		
		for(int i = 0; i < myChildList.length; ++i) {
		    LinkedList myVeryTempList = myChildList[i].getRelationships(inCrossThisPackage);
		    
		    if(myVeryTempList != null)
			myTempList.addAll(myVeryTempList);
		}
		
		return myTempList;
	}
    
	public void addRelationship(VRelationship inRelationship)
	{
		if(mOriginalRelationships == null)
			mOriginalRelationships = new LinkedList();

		if(inRelationship.getSource() == this ||
		        inRelationship.getDestination() == this)
			mOriginalRelationships.add(inRelationship);
	}

	// VModelElement operations end

	public VRelationship addToCompositeRelationshipList(VRelationship inRel)
	{
		boolean isSourceInThisPackage;

		if(((ZNode) inRel.getSource()).isDescendentOf(this))
			isSourceInThisPackage = true;
		else
			isSourceInThisPackage = false;

		inRel.setIsRepresentedBySomeoneElse(true);
		ListIterator myIterator = mCompositeRelationships.listIterator(0);

		while(myIterator.hasNext())
		{
			ZRelationship myZRelationship = (ZRelationship) myIterator.next();

			// Source is in this package, so only type, relationship type
			// and destination have to match. Source doesn't have to match,
			// because source is this in the composite relationship.
			if(isSourceInThisPackage)
			{
				if(myZRelationship.getType() == inRel.getType() &&
				        myZRelationship.getDestination() == inRel.getDestination()
				        && myZRelationship.getRelationshipType() ==
				        ((ZRelationship) inRel).getRelationshipType())
				{
					myZRelationship.addVRelationshipAsSubrelationship(inRel);
					return null;
				}
			}
			else
			{
				if(myZRelationship.getType() == inRel.getType() &&
				        myZRelationship.getSource() == inRel.getSource() &&
				        myZRelationship.getRelationshipType() ==
				        ((ZRelationship) inRel).getRelationshipType())
				{
					myZRelationship.addVRelationshipAsSubrelationship(inRel);
					return null;
				}

			}
		}

		// No matching composite relationship found, must create a new one.

		ZRelationship myZRelationship = null;

		// FIXME!! Use ZVisualization to create an unique ID
		if(inRel instanceof ZRealization)
			myZRelationship = new ZRealization(null, mListenersEnabled);
		else if(inRel instanceof ZGeneralization)
			myZRelationship = new ZGeneralization(null, mListenersEnabled);
		else if(inRel instanceof ZDependency)
			myZRelationship = new ZDependency(null, mListenersEnabled);

		myZRelationship.setType(inRel.getType());
		myZRelationship.setName("Composite relationship");

		if(isSourceInThisPackage)
		{
			myZRelationship.setSource(this);
			myZRelationship.setDestination(inRel.getDestination());
		}
		else
		{
			myZRelationship.setSource(inRel.getSource());
			myZRelationship.setDestination(this);
		}

		myZRelationship.addVRelationshipAsSubrelationship(inRel);
		
		mCompositeRelationships.add(myZRelationship);
		
		ZVisualization myVis = ZVisualization.getZVisualizationInstance();
		
		myVis.addVRelationshipWithoutAddingOriginalReference(myZRelationship);
		
		return myZRelationship;
	}

	public LinkedList getCompositeRelationships()
	{
		return mCompositeRelationships;
	}

    public void setState(int inState)
    {
	if(inState == OPEN || inState == CLOSED || inState == TRANSPARENT)
	    mState = inState;
	else
	    System.err.println("Tried to set package state to invalid value " + inState);
    }
    
    public int getState()
    {
	return mState;
    }
    
    public void debugDump()
    {
	VRelationship myRel = null;
	
	ListIterator it = mCompositeRelationships.listIterator();
	
	while(it.hasNext()) {
	    myRel = (VRelationship) it.next();
	    System.out.println("comp rel src " + myRel.getSource().getId() + " dst " + myRel.getDestination().getId() + " " + myRel.getIsRepresentedBySomeoneElse());
	}
    }
}
