/*
 * @(#) ZRelationship.java
 */

// TODO: don't draw the relationship if
//  - target/source not in scenegraph

package venice;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.component.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;

import java.awt.*;
import java.awt.geom.*;
import java.util.LinkedList;
import java.util.ListIterator;

import javax.swing.SwingUtilities;

/**
 * ZRelationship is an implementation of VRelationship interface using
 * Jazz graphics library. It can't be instantiated directly but you
 * can instantiate one of its subclasses which are just wrappers
 * for ZRelationship
 * <p>
 * The arrowhead consist of three lines in the following way:
 * <pre>
 *
 *   |\ 2 
 *   | \
 * 1 | /
 *   |/ 3
 *
 * 1) mArrowHeadUpDownLine
 * 2) mArrowHeadUpLeftLine
 * 3) mArrowHeadDownLeftLine
 * </pre>
 *
 * @author Hannu Laurila
 */
public abstract class ZRelationship extends ZVisualLeafModelElement
    implements VRelationship, ZNodeListener, ListenerStateChangeable
{
    /** a constant which tells that this instance is a dependency */
    protected static final int TYPE_DEPENDENCY = 1;
    
    /** a constant which tells that this instance is a generalization */
    protected static final int TYPE_GENERALIZATION = 2;
    
    /** a constant which tells that this instance is a realization */
    protected static final int TYPE_REALIZATION = 3;
    
    /** pixels used to draw the arrowheads. Increates to get bigger arrows */
    protected static final double HEADLEN = 6.0;
    
    /** source model element of the relationship */
    private ZNode mSrc;	
    
    /** destination model element of the relationship */
    private ZNode mDst;
    
    /** visual component used to draw the line from src to dest */
    private ZLine mLine;

    /** visual components used to draw the arrowhead to the line */
    private ZLine mArrowHeadUpLeftLine;
    private ZLine mArrowHeadDownLeftLine;
    private ZLine mArrowHeadUpDownLine;
    
    /** visual component used to draw the label on the line */
    private ZLabel mTypeLabel;
    
    /** type of this relationship. One of TYPE_ constants */
    private int mRelationshipType;

    /** in case of realization, this variable stores last drawing style  */
    private boolean mPreviousUseDashed;
    
    /** The list of relationships which are represented by this relationship */
    private LinkedList mCompositeRelationshipList;
    
    /** Is this relationship represented by some other relationship */
    private boolean mIsRepresentedByOtherRelationship;
	
    /** a Java2D pen stroke used to draw a dashed line (e.g. dependency) */
    private static BasicStroke stDashedStroke = 
	new BasicStroke(1,
			BasicStroke.CAP_BUTT,
			BasicStroke.JOIN_BEVEL,
			0,   // Miter limit
			new float[] {4,4},
			0);  // Dash phase

    /** a Java2D pen stroke use to draw a solid line **/
    private static BasicStroke stSolidStroke =
        new BasicStroke(1, 
                        BasicStroke.CAP_BUTT, 
                        BasicStroke.JOIN_MITER);
    
    /**
     * a not-serialized variable telling that no listeners are enabled.
     * after deserialization, the value of this will always be false.
     */
    private transient boolean mListenersEnabled = false;
    
    /**
     * constructor used by subclasses to instantiate a ZRelationship
     * with corrent relationship type
     */
    public ZRelationship(String inId, int inRelationshipType, 
                         boolean inInitialListenersEnabled)
    {
	super(inId);
	
	this.mSrc = null;
	this.mDst = null;
	
	this.mRelationshipType = inRelationshipType;
	this.mPreviousUseDashed = false;

	this.mLine = new ZLine();
	this.mTypeLabel = new ZLabel();
	this.mArrowHeadUpLeftLine = new ZLine();
	this.mArrowHeadDownLeftLine = new ZLine();
	this.mArrowHeadUpDownLine = new ZLine();
	
	this.setSelectable(false);
        
	this.mListenersEnabled = inInitialListenersEnabled;

	this.mCompositeRelationshipList = null;
	this.mIsRepresentedByOtherRelationship = false;
	
	// add the all visual components which are common to all relships
	addVisualComponent(mLine);
	//addVisualComponent(mTypeLabel);
	
	/*
	 * determine by relationship type if to have the up->down line
	 * part of the arrowat all and whether to use dashed or
	 * normal stroke. If mRelationshipType is not recognized,
	 * it is considered to be a dependency.
	 */
	switch (mRelationshipType)
	{
	    case TYPE_GENERALIZATION:
		addVisualComponent(mArrowHeadUpLeftLine);
		addVisualComponent(mArrowHeadDownLeftLine);
		addVisualComponent(mArrowHeadUpDownLine);
		break;
	    case TYPE_REALIZATION:
		// FIXED: 9.4.2001 Petteri Kamppuri commented these out
                // Now dst VIEWMODE_CIRLCE is assumed by default
		//mLine.setStroke(stDashedStroke);
		//addVisualComponent(mArrowHeadUpLeftLine);
		//addVisualComponent(mArrowHeadDownLeftLine);
		//addVisualComponent(mArrowHeadUpDownLine);
		break;
	    case TYPE_DEPENDENCY:
	    default:
		mLine.setStroke(stDashedStroke);
		mRelationshipType = TYPE_DEPENDENCY;
		addVisualComponent(mArrowHeadUpLeftLine);
		addVisualComponent(mArrowHeadDownLeftLine);
		break;
	}
	
	dataChanged();
	
    }
    
    public VModelElement getSource()
    {
	return (VModelElement) mSrc;
    }
    
    public VModelElement getDestination() 
    {
	return (VModelElement) mDst;
    }
    
    public void setSource(VModelElement inElement)
    {
	ZNode inNode = (ZNode) inElement;
	
	if (inNode == mSrc) {
	    // no change
	    return;
	}
	
	// remove listener for current source if there is current source
	
        if (mListenersEnabled)
            if (mSrc != null)
                mSrc.removeNodeListener(this);
	
	// change the source and if it's not null, start listening
	// for bounds changes.
	
	mSrc = inNode;
	
        if (mListenersEnabled)
            if (mSrc != null)
                mSrc.addNodeListener(this);
	
	dataChanged();
    }
    
    public void setDestination(VModelElement inElement)
    {
	ZNode inNode = (ZNode) inElement;
	
	if (inNode == mDst) {
	    // no change
	    return;
	}
	
	// remove listener for current source if there is current source
	
        if (mListenersEnabled)
            if (mDst != null)
                mDst.removeNodeListener(this);
	
	// change the source and if it's not null, start listening
	// for bounds changes.
	
	mDst = inNode;
	
        if (mListenersEnabled)
            if (mDst != null)
                mDst.addNodeListener(this);
	
	dataChanged();
    }
    
    public Point2D getRelationshipConnectPoint()
    {
	ZBounds myBounds = getGlobalBounds();
	
	return myBounds.getCenter2D();
    }


    public void setListenersEnabled(boolean inEnabled, boolean inRecurse)
    {
        if (mListenersEnabled == true && inEnabled == false) {
            if (mDst != null)
                mDst.removeNodeListener(this);
            if (mSrc != null)
                mSrc.removeNodeListener(this);
        }
        
        if (mListenersEnabled == false && inEnabled == true) {
            if (mDst != null)
                mDst.addNodeListener(this);
            if (mSrc != null)
                mSrc.addNodeListener(this);
        }
        
        mListenersEnabled = inEnabled;

        dataChanged();
    }
    
    /**
     * takes care of adjusting the appearance according to changes
     * in information
     */
    protected void dataChanged()
    {
        // source point of the relationship in local coordinates
        Point2D mSrcPoint;	
        
        // destination point of the relationship in local coordinates
        Point2D mDstPoint;

	if (mModelElementVisible == false)
	    this.editor().getInvisibleGroup();
	else
	    this.editor().removeInvisibleGroup();
	
	if (mSrc == null || mDst == null) {
	    // Source and target have not been set yet - no need
	    // to setup lines, arrowheadsetc. yet.
	    //
	    // FIXME: if src or dst is modified to be null after
	    // they have once been something useful, lines stay
	    // in the screen which is wrong
	    return;
	}

	// Uncomment the following block of code to automatically hide
	// relationships, whose source or destination is invisible.
	// This is not something we need, so it is not done.
	/*if(getSource().getVisible() == false || getDestination().getVisible() == false) {
	    this.editor().getInvisibleGroup();
	    return;
	    }*/
	
	/*
	 * we have to work using global coordinates firt because their
	 * space is common to every element in the visualization.
	 * After we have
	 * found out the places in global coordinates, we can translate
	 * them back to THIS nodes local coordinates.
	 */
	
	VModelElement mySource = (VModelElement) mSrc;
	VModelElement myDestination = (VModelElement) mDst;
	
	// create a 2D line from source center to destination center
	Line2D myLine = new Line2D.Double(mySource.getRelationshipConnectPoint(),
					  myDestination.getRelationshipConnectPoint());
	
	// calculate the intersection points for the line and the
	// element bounds and use them as source and destination point.
	// This basically means that the lines are shortened so
	// that they don't cross the element's bounds.
	
	// mSrc and mDst are always somekinds of VModelElements.
	// If they are not, then there is a fault somewhere in the
	// code. It is a clear error, and it should be fixed.
	
	mSrcPoint = mySource.intersectWithLine(myLine);
	
	myLine = new Line2D.Double(myDestination.getRelationshipConnectPoint(),
				   mySource.getRelationshipConnectPoint());
	
	mDstPoint = myDestination.intersectWithLine(myLine);
	
	// if either source or target did not intersect with the
	// bounds rectangle, use centers as the source and target
	// coordinates
	if (mSrcPoint == null || mDstPoint == null) {
	    mSrcPoint = mySource.getRelationshipConnectPoint();
	    mDstPoint = myDestination.getRelationshipConnectPoint();
	}
	
	// move to using local coordinates (local to ZRelationship)
	// and draw the line.
	globalToLocal(mSrcPoint);
	globalToLocal(mDstPoint);
	mLine.setLine(mSrcPoint, mDstPoint);
	
	// set up the label and position it on about the center of the line
	mTypeLabel.setText(mModelElementType);
	
	double myNameLabelX = 
	    ((mSrcPoint.getX() + mDstPoint.getX()) / 2) - 
	    mTypeLabel.getBoundsReference().getWidth()/2;
	
	double myNameLabelY = 
	    ((mSrcPoint.getY() + mDstPoint.getY()) / 2) - 20;
	
	mTypeLabel.setTranslation(myNameLabelX, myNameLabelY);

	/* 
         * Hack: Prepare arrowhead in case we are drawing a realization.
         * If destination is a VInterface which is in VIEWMODE_EXPANDED
         * mode, use arrow with dashed stroke and arrowhead. Otherwise
         * use solid stroke and no arrowhead.
         */
        
        boolean useDashed = true;

        if (mRelationshipType == TYPE_REALIZATION) {
            if (mDst instanceof VInterface && 
                ( ((VInterface)mDst).getViewMode() == VInterface.VIEWMODE_CIRCLE))
                useDashed = false;

            // previous useDashed is cached to prevent unnecessary
            // adding and removing of visual components
            if (useDashed != mPreviousUseDashed) {
                mPreviousUseDashed = useDashed;            
                removeVisualComponent(mArrowHeadUpLeftLine);
                removeVisualComponent(mArrowHeadDownLeftLine);
                removeVisualComponent(mArrowHeadUpDownLine);
                
                if (useDashed) {
                    mLine.setStroke(stDashedStroke);
                    addVisualComponent(mArrowHeadUpLeftLine);
                    addVisualComponent(mArrowHeadDownLeftLine);
                    addVisualComponent(mArrowHeadUpDownLine);
                } else {
                    mLine.setStroke(stSolidStroke);
                }
            }
        }
        
	/*
	 * draw arrowhead. This is a bit complicated and slow
	 */
	// FIXED: 9.4.2001 Petteri Kamppuri added check to prevent
	// unnecessary heavy calculations
	if( (mRelationshipType != TYPE_REALIZATION) || (useDashed == true)) {
	    // find out the angle of the line
	    double myTheta = Math.atan2(mDstPoint.getY() - mSrcPoint.getY(),
					mDstPoint.getX() - mSrcPoint.getX());
	    
	    // create a transform which rotates points relative to the
	    // destination point accroding to the angle of the line.
	    AffineTransform myAT = 
		AffineTransform.getRotateInstance(myTheta,
						  mDstPoint.getX(),
						  mDstPoint.getY());
	    
	    Point2D myTmpPoint = null;
	    Point2D myUpLeftPoint = new Point2D.Double();
	    Point2D myDownLeftPoint = new Point2D.Double();
	    Point2D myLeftPoint = new Point2D.Double();
	    
	    /* up left point */	
	    myTmpPoint = new Point2D.Double(mDstPoint.getX() - HEADLEN,
					    mDstPoint.getY() + HEADLEN);
	    myAT.transform(myTmpPoint, myUpLeftPoint);
	    
	    /* down left point */
	    
	    myTmpPoint = new Point2D.Double(mDstPoint.getX() - HEADLEN,
					    mDstPoint.getY() - HEADLEN);
	    myAT.transform(myTmpPoint, myDownLeftPoint);
	    
	    /* left point */
	    
	    myTmpPoint = new Point2D.Double(mDstPoint.getX() - HEADLEN,
					    mDstPoint.getY());
	    myAT.transform(myTmpPoint, myLeftPoint);
	    
	    
	    /* draw the lines */
	    
	    mArrowHeadUpLeftLine.setLine(mDstPoint, myUpLeftPoint);
	    mArrowHeadDownLeftLine.setLine(mDstPoint, myDownLeftPoint);
	    mArrowHeadUpDownLine.setLine(myUpLeftPoint, myDownLeftPoint);
	    
	    // if this is not a dependency, the line should not be drawn
	    // to target point but to the center of the up-down line.
	    if (mRelationshipType != TYPE_DEPENDENCY)
		mLine.setLine(mSrcPoint, myLeftPoint);
	}
    }
    
    public void boundsChanged(ZNodeEvent e) 
    {
	// We are not interested in these events
    }
    
    public void globalBoundsChanged(ZNodeEvent e)
    {
	// either source or destination node has moved.
	dataChanged();
    }
    
    public Point2D intersectWithLine(Line2D inLine)
    {
	return null;
    }
    
    /**
     * As relationships cannot have relationships
     * attached to themselves, this method will always return null.
     * 
     * @return always returns null
     */
    public LinkedList getRelationships(VPackage inLeaveThisPackage)
    {
	return null;
    }
    
    /**
     * Relationships cannot have relationships attached to themselves
     * so this method does nothing.
     */
    public void addRelationship(VRelationship inRelationship)
    {
	System.err.println("addRelationship should never be called for ZRelationship");
    }
    
    public boolean leavesVPackage(VPackage inPackage)
    {
	if(inPackage == null)
	    return true;
	ZGroup myGroup = (ZGroup) inPackage;
	
	if(((ZNode) getSource()).isDescendentOf(myGroup) &&
	   ((ZNode) getDestination()).isDescendentOf(myGroup))
	    return false;
	else
	    return true;
    }
    
    public void addVRelationshipAsSubrelationship(VRelationship inRel)
    {
	if(mCompositeRelationshipList == null)
	    mCompositeRelationshipList = new LinkedList();
	
	mCompositeRelationshipList.add(inRel);
	//ZVisualization.getZVisualizationInstance().removeVRelationship(inRel);
    }
    
    public boolean isCompositeRelationship()
    {
	return mCompositeRelationshipList != null;
    }
    
    public LinkedList getSubrelationships()
    {
	return mCompositeRelationshipList;
    }
    
    public void removeSubrelationship(VRelationship inRelationshipToRemove)
    {
	inRelationshipToRemove.setIsRepresentedBySomeoneElse(false);
	//mCompositeRelationshipList.remove(inRelationshipToRemove);
    }
    
    /**
     * Returns the type of this relationship which may be one of the following:
     * relization, generalization or dependency.
     * <p>This is a partial hack solution, because one might argue that
     * type should be represented as a class hierarchy.
     * 
     * @return the type of this relationship
     */
    public int getRelationshipType()
    {
	return mRelationshipType;
    }
    
    public void setIsRepresentedBySomeoneElse(boolean inValue)
    {
	mIsRepresentedByOtherRelationship = inValue;
    }
    
    public boolean getIsRepresentedBySomeoneElse()
    {
	return mIsRepresentedByOtherRelationship;
    }

    public void setLabelVisible(boolean inLabelVisible)
    {
	if(inLabelVisible) {
	    addVisualComponent(mTypeLabel);
	} else {
	    removeVisualComponent(mTypeLabel);
	}
    }
}
